from django.http import HttpRequest, JsonResponse
from django.utils.html import escape
from django.utils.translation import gettext as _

from debug_toolbar._compat import login_not_required
from debug_toolbar.decorators import render_with_toolbar_language, require_show_toolbar
from debug_toolbar.panels import Panel
from debug_toolbar.toolbar import DebugToolbar, StoredDebugToolbar


@login_not_required
@require_show_toolbar
@render_with_toolbar_language
def render_panel(request: HttpRequest) -> JsonResponse:
    """Render the contents of a panel"""
    toolbar: StoredDebugToolbar | None = DebugToolbar.fetch(
        request.GET["request_id"], request.GET["panel_id"]
    )
    if toolbar is None:
        content = _(
            "Data for this panel isn't available anymore. "
            "Please reload the page and retry."
        )
        content = f"<p>{escape(content)}</p>"
        scripts = []
    else:
        panel: Panel = toolbar.get_panel_by_id(request.GET["panel_id"])
        content = panel.content
        scripts = panel.scripts
    return JsonResponse({"content": content, "scripts": scripts})



"""urls.py to test using debug_toolbar.urls in include"""

from django.urls import include, path

import debug_toolbar

from . import views

urlpatterns = [
    path("cached_view/", views.cached_view),
    path("__debug__/", include(debug_toolbar.urls)),
]



from django.core.cache import cache


class UseCacheAfterToolbar:
    """
    This middleware exists to use the cache before and after
    the toolbar is setup.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        cache.set("UseCacheAfterToolbar.before", 1)
        response = self.get_response(request)
        cache.set("UseCacheAfterToolbar.after", 1)
        return response



import uuid
from collections import defaultdict

from asgiref.sync import sync_to_async
from django.db import connections
from django.template.loader import render_to_string
from django.urls import path
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _, ngettext

from debug_toolbar import settings as dt_settings
from debug_toolbar.forms import SignedDataForm
from debug_toolbar.panels import Panel
from debug_toolbar.panels.sql import views
from debug_toolbar.panels.sql.forms import SQLSelectForm
from debug_toolbar.panels.sql.tracking import wrap_cursor
from debug_toolbar.panels.sql.utils import (
    contrasting_color_generator,
    reformat_sql,
)
from debug_toolbar.utils import render_stacktrace


def get_isolation_level_display(vendor, level):
    if vendor == "postgresql":
        try:
            import psycopg

            choices = {
                # AUTOCOMMIT level does not exists in psycopg3
                psycopg.IsolationLevel.READ_UNCOMMITTED: _("Read uncommitted"),
                psycopg.IsolationLevel.READ_COMMITTED: _("Read committed"),
                psycopg.IsolationLevel.REPEATABLE_READ: _("Repeatable read"),
                psycopg.IsolationLevel.SERIALIZABLE: _("Serializable"),
            }
        except ImportError:
            import psycopg2.extensions

            choices = {
                psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT: _("Autocommit"),
                psycopg2.extensions.ISOLATION_LEVEL_READ_UNCOMMITTED: _(
                    "Read uncommitted"
                ),
                psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED: _("Read committed"),
                psycopg2.extensions.ISOLATION_LEVEL_REPEATABLE_READ: _(
                    "Repeatable read"
                ),
                psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE: _("Serializable"),
            }

    else:
        raise ValueError(vendor)
    return choices.get(level)


def get_transaction_status_display(vendor, level):
    if vendor == "postgresql":
        try:
            import psycopg

            choices = {
                psycopg.pq.TransactionStatus.IDLE: _("Idle"),
                psycopg.pq.TransactionStatus.ACTIVE: _("Active"),
                psycopg.pq.TransactionStatus.INTRANS: _("In transaction"),
                psycopg.pq.TransactionStatus.INERROR: _("In error"),
                psycopg.pq.TransactionStatus.UNKNOWN: _("Unknown"),
            }
        except ImportError:
            import psycopg2.extensions

            choices = {
                psycopg2.extensions.TRANSACTION_STATUS_IDLE: _("Idle"),
                psycopg2.extensions.TRANSACTION_STATUS_ACTIVE: _("Active"),
                psycopg2.extensions.TRANSACTION_STATUS_INTRANS: _("In transaction"),
                psycopg2.extensions.TRANSACTION_STATUS_INERROR: _("In error"),
                psycopg2.extensions.TRANSACTION_STATUS_UNKNOWN: _("Unknown"),
            }

    else:
        raise ValueError(vendor)
    return choices.get(level)


def _similar_query_key(query):
    return query["raw_sql"]


def _duplicate_query_key(query):
    raw_params = () if query["params"] is None else tuple(query["params"])
    # repr() avoids problems because of unhashable types
    # (e.g. lists) when used as dictionary keys.
    # https://github.com/django-commons/django-debug-toolbar/issues/1091
    return (query["raw_sql"], repr(raw_params))


def _process_query_groups(query_groups, databases, colors, name):
    counts = defaultdict(int)
    for (alias, _key), query_group in query_groups.items():
        count = len(query_group)
        # Queries are similar / duplicates only if there are at least 2 of them.
        if count > 1:
            color = next(colors)
            for query in query_group:
                query[f"{name}_count"] = count
                query[f"{name}_color"] = color
            counts[alias] += count
    for alias, db_info in databases.items():
        db_info[f"{name}_count"] = counts[alias]


class SQLPanel(Panel):
    """
    Panel that displays information about the SQL queries run while processing
    the request.
    """

    is_async = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sql_time = 0
        self._queries = []
        self._databases = {}
        # synthetic transaction IDs, keyed by DB alias
        self._transaction_ids = {}

    def new_transaction_id(self, alias):
        """
        Generate and return a new synthetic transaction ID for the specified DB alias.
        """
        trans_id = uuid.uuid4().hex
        self._transaction_ids[alias] = trans_id
        return trans_id

    def current_transaction_id(self, alias):
        """
        Return the current synthetic transaction ID for the specified DB alias.
        """
        trans_id = self._transaction_ids.get(alias)
        # Sometimes it is not possible to detect the beginning of the first transaction,
        # so current_transaction_id() will be called before new_transaction_id().  In
        # that case there won't yet be a transaction ID. so it is necessary to generate
        # one using new_transaction_id().
        if trans_id is None:
            trans_id = self.new_transaction_id(alias)
        return trans_id

    def record(self, **kwargs):
        kwargs["djdt_query_id"] = uuid.uuid4().hex
        self._queries.append(kwargs)
        alias = kwargs["alias"]
        if alias not in self._databases:
            self._databases[alias] = {
                "time_spent": kwargs["duration"],
                "num_queries": 1,
            }
        else:
            self._databases[alias]["time_spent"] += kwargs["duration"]
            self._databases[alias]["num_queries"] += 1
        self._sql_time += kwargs["duration"]

    # Implement the Panel API

    nav_title = _("SQL")
    nav_icon = "debug_toolbar/img/sql.svg"

    @property
    def nav_subtitle(self):
        stats = self.get_stats()
        query_count = len(stats.get("queries", []))
        return ngettext(
            "%(query_count)d query in %(sql_time).2fms",
            "%(query_count)d queries in %(sql_time).2fms",
            query_count,
        ) % {
            "query_count": query_count,
            "sql_time": stats.get("sql_time"),
        }

    @property
    def title(self):
        count = len(self.get_stats().get("databases"))
        return ngettext(
            "SQL queries from %(count)d connection",
            "SQL queries from %(count)d connections",
            count,
        ) % {"count": count}

    template = "debug_toolbar/panels/sql.html"

    @classmethod
    def get_urls(cls):
        return [
            path("sql_select/", views.sql_select, name="sql_select"),
            path("sql_explain/", views.sql_explain, name="sql_explain"),
            path("sql_profile/", views.sql_profile, name="sql_profile"),
        ]

    async def aenable_instrumentation(self):
        """
        Async version of enable instrumentation.
        For async capable panels having async logic for instrumentation.
        """
        await sync_to_async(self.enable_instrumentation)()

    def enable_instrumentation(self):
        # This is thread-safe because database connections are thread-local.
        for connection in connections.all():
            wrap_cursor(connection)
            connection._djdt_logger = self

    def disable_instrumentation(self):
        for connection in connections.all():
            connection._djdt_logger = None

    def generate_stats(self, request, response):
        similar_query_groups = defaultdict(list)
        duplicate_query_groups = defaultdict(list)

        if self._queries:
            sql_warning_threshold = dt_settings.get_config()["SQL_WARNING_THRESHOLD"]

            width_ratio_tally = 0
            factor = int(256.0 / (len(self._databases) * 2.5))
            for n, db in enumerate(self._databases.values()):
                rgb = [0, 0, 0]
                color = n % 3
                rgb[color] = 256 - n // 3 * factor
                nn = color
                # XXX: pretty sure this is horrible after so many aliases
                while rgb[color] < factor:
                    nc = min(256 - rgb[color], 256)
                    rgb[color] += nc
                    nn += 1
                    if nn > 2:
                        nn = 0
                    rgb[nn] = nc
                db["rgb_color"] = rgb

            # the last query recorded for each DB alias
            last_by_alias = {}
            for query in self._queries:
                alias = query["alias"]

                similar_query_groups[(alias, _similar_query_key(query))].append(query)
                duplicate_query_groups[(alias, _duplicate_query_key(query))].append(
                    query
                )

                trans_id = query.get("trans_id")
                prev_query = last_by_alias.get(alias, {})
                prev_trans_id = prev_query.get("trans_id")

                # If two consecutive queries for a given DB alias have different
                # transaction ID values, a transaction started, finished, or both, so
                # annotate the queries as appropriate.
                if trans_id != prev_trans_id:
                    if prev_trans_id is not None:
                        prev_query["ends_trans"] = True
                    if trans_id is not None:
                        query["starts_trans"] = True
                if trans_id is not None:
                    query["in_trans"] = True

                if "iso_level" in query:
                    query["iso_level"] = get_isolation_level_display(
                        query["vendor"], query["iso_level"]
                    )
                if "trans_status" in query:
                    query["trans_status"] = get_transaction_status_display(
                        query["vendor"], query["trans_status"]
                    )
                query["is_slow"] = query["duration"] > sql_warning_threshold

                query["rgb_color"] = self._databases[alias]["rgb_color"]
                try:
                    query["width_ratio"] = (query["duration"] / self._sql_time) * 100
                except ZeroDivisionError:
                    query["width_ratio"] = 0
                query["start_offset"] = width_ratio_tally
                query["end_offset"] = query["width_ratio"] + query["start_offset"]
                width_ratio_tally += query["width_ratio"]

                last_by_alias[alias] = query

            # Close out any transactions that were in progress, since there is no
            # explicit way to know when a transaction finishes.
            for final_query in last_by_alias.values():
                if final_query.get("trans_id") is not None:
                    final_query["ends_trans"] = True

        group_colors = contrasting_color_generator()
        _process_query_groups(
            similar_query_groups, self._databases, group_colors, "similar"
        )
        _process_query_groups(
            duplicate_query_groups, self._databases, group_colors, "duplicate"
        )

        self.record_stats(
            {
                "databases": sorted(
                    self._databases.items(), key=lambda x: -x[1]["time_spent"]
                ),
                "queries": self._queries,
                "sql_time": self._sql_time,
            }
        )

    def generate_server_timing(self, request, response):
        stats = self.get_stats()
        title = "SQL {} queries".format(len(stats.get("queries", [])))
        value = stats.get("sql_time", 0)
        self.record_server_timing("sql_time", title, value)

    # Cache the content property since it manipulates the queries in the stats
    # This allows the caller to treat content as idempotent
    @cached_property
    def content(self):
        if self.has_content:
            stats = self.get_stats()
            colors = contrasting_color_generator()
            trace_colors = defaultdict(lambda: next(colors))

            for query in stats.get("queries", []):
                query["sql"] = reformat_sql(query["sql"], with_toggle=True)
                query["form"] = SignedDataForm(
                    auto_id=None,
                    initial=SQLSelectForm(
                        initial={
                            "djdt_query_id": query["djdt_query_id"],
                            "request_id": self.toolbar.request_id,
                        }
                    ).initial,
                )
                query["stacktrace"] = render_stacktrace(query["stacktrace"])
                query["trace_color"] = trace_colors[query["stacktrace"]]

            return render_to_string(self.template, stats)



from unittest import expectedFailure

from django.contrib.auth.models import User
from django.template import Context, RequestContext, Template
from django.test import override_settings
from django.utils.functional import SimpleLazyObject

from debug_toolbar.panels.sql import SQLPanel
from debug_toolbar.panels.templates import TemplatesPanel

from ..base import BaseTestCase, IntegrationTestCase
from ..forms import TemplateReprForm
from ..models import NonAsciiRepr


class TemplatesPanelTestCase(BaseTestCase):
    panel_id = TemplatesPanel.panel_id

    def setUp(self):
        super().setUp()
        self.sql_panel = self.toolbar.get_panel_by_id(SQLPanel.panel_id)
        self.sql_panel.enable_instrumentation()

    def tearDown(self):
        self.sql_panel.disable_instrumentation()
        super().tearDown()

    def test_queryset_hook(self):
        response = self.panel.process_request(self.request)
        t = Template("No context variables here!")
        c = Context(
            {
                "queryset": User.objects.all(),
                "deep_queryset": {"queryset": User.objects.all()},
            }
        )
        t.render(c)
        self.panel.generate_stats(self.request, response)

        # ensure the query was NOT logged
        self.assertEqual(len(self.sql_panel._queries), 0)

        self.assertEqual(
            self.panel.templates[0]["context_list"],
            [
                "{'False': False, 'None': None, 'True': True}",
                (
                    "{'deep_queryset': '<<triggers database query>>',\n"
                    " 'queryset': '<<queryset of auth.User>>'}"
                ),
            ],
        )

    def test_template_repr(self):
        # Force widget templates to be included
        self.toolbar.config["SKIP_TEMPLATE_PREFIXES"] = ()

        User.objects.create(username="admin")
        bad_repr = TemplateReprForm()
        t = Template("{{ bad_repr }}")
        c = Context({"bad_repr": bad_repr})
        html = t.render(c)
        self.assertIsNotNone(html)
        self.assertValidHTML(html)

    def test_object_with_non_ascii_repr_in_context(self):
        response = self.panel.process_request(self.request)
        t = Template("{{ object }}")
        c = Context({"object": NonAsciiRepr()})
        t.render(c)
        self.panel.generate_stats(self.request, response)
        self.assertIn("nôt åscíì", self.panel.content)

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        t = Template("{{ object }}")
        c = Context({"object": NonAsciiRepr()})
        t.render(c)
        response = self.panel.process_request(self.request)
        # ensure the panel does not have content yet.
        self.assertNotIn("nôt åscíì", self.panel.content)
        self.panel.generate_stats(self.request, response)
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn("nôt åscíì", content)
        self.assertValidHTML(content)

    def test_custom_context_processor(self):
        response = self.panel.process_request(self.request)
        t = Template("{{ content }}")
        c = RequestContext(self.request, processors=[context_processor])
        t.render(c)
        self.panel.generate_stats(self.request, response)
        self.assertIn(
            "tests.panels.test_template.context_processor", self.panel.content
        )

    def test_disabled(self):
        config = {"DISABLE_PANELS": {"debug_toolbar.panels.templates.TemplatesPanel"}}
        self.assertTrue(self.panel.enabled)
        with self.settings(DEBUG_TOOLBAR_CONFIG=config):
            self.assertFalse(self.panel.enabled)

    def test_empty_context(self):
        response = self.panel.process_request(self.request)
        t = Template("")
        c = Context({})
        t.render(c)
        self.panel.generate_stats(self.request, response)

        # Includes the builtin context but not the empty one.
        self.assertEqual(
            self.panel.templates[0]["context_list"],
            ["{'False': False, 'None': None, 'True': True}"],
        )

    def test_lazyobject(self):
        response = self.panel.process_request(self.request)
        t = Template("")
        c = Context({"lazy": SimpleLazyObject(lambda: "lazy_value")})
        t.render(c)
        self.panel.generate_stats(self.request, response)
        self.assertNotIn("lazy_value", self.panel.content)

    def test_lazyobject_eval(self):
        response = self.panel.process_request(self.request)
        t = Template("{{lazy}}")
        c = Context({"lazy": SimpleLazyObject(lambda: "lazy_value")})
        self.assertEqual(t.render(c), "lazy_value")
        self.panel.generate_stats(self.request, response)
        self.assertIn("lazy_value", self.panel.content)

    @override_settings(
        DEBUG=True,
        DEBUG_TOOLBAR_PANELS=["debug_toolbar.panels.templates.TemplatesPanel"],
    )
    def test_template_source(self):
        from django.core import signing
        from django.template.loader import get_template

        template = get_template("basic.html")
        url = "/__debug__/template_source/"
        data = {
            "template": template.template.name,
            "template_origin": signing.dumps(template.template.origin.name),
        }

        response = self.client.get(url, data)
        self.assertEqual(response.status_code, 200)


@override_settings(
    DEBUG=True, DEBUG_TOOLBAR_PANELS=["debug_toolbar.panels.templates.TemplatesPanel"]
)
class JinjaTemplateTestCase(IntegrationTestCase):
    def test_django_jinja2(self):
        r = self.client.get("/regular_jinja/foobar/")
        self.assertContains(r, "Test for foobar (Jinja)")
        # This should be 2 templates because of the parent template.
        # See test_django_jinja2_parent_template_instrumented
        self.assertContains(r, "<h3>Templates (1 rendered)</h3>")
        self.assertContains(r, "<small>basic.jinja</small>")

    @expectedFailure
    def test_django_jinja2_parent_template_instrumented(self):
        """
        When Jinja2 templates are properly instrumented, the
        parent template should be instrumented.
        """
        r = self.client.get("/regular_jinja/foobar/")
        self.assertContains(r, "Test for foobar (Jinja)")
        self.assertContains(r, "<h3>Templates (2 rendered)</h3>")
        self.assertContains(r, "<small>basic.jinja</small>")


def context_processor(request):
    return {"content": "set by processor"}



import copy

from django.conf import settings
from django.http import HttpResponse
from django.test import AsyncRequestFactory

from debug_toolbar.panels.redirects import RedirectsPanel

from ..base import BaseTestCase


class RedirectsPanelTestCase(BaseTestCase):
    panel_id = RedirectsPanel.panel_id

    def test_regular_response(self):
        not_redirect = HttpResponse()
        self._get_response = lambda request: not_redirect
        response = self.panel.process_request(self.request)
        self.assertTrue(response is not_redirect)

    def test_not_a_redirect(self):
        redirect = HttpResponse(status=304)
        self._get_response = lambda request: redirect
        response = self.panel.process_request(self.request)
        self.assertTrue(response is redirect)

    def test_redirect(self):
        redirect = HttpResponse(status=302)
        redirect["Location"] = "http://somewhere/else/"
        self._get_response = lambda request: redirect
        response = self.panel.process_request(self.request)
        self.assertFalse(response is redirect)
        self.assertContains(response, "302 Found")
        self.assertContains(response, "http://somewhere/else/")

    def test_redirect_with_broken_context_processor(self):
        TEMPLATES = copy.deepcopy(settings.TEMPLATES)
        TEMPLATES[1]["OPTIONS"]["context_processors"] = [
            "tests.context_processors.broken"
        ]

        with self.settings(TEMPLATES=TEMPLATES):
            redirect = HttpResponse(status=302)
            redirect["Location"] = "http://somewhere/else/"
            self._get_response = lambda request: redirect
            response = self.panel.process_request(self.request)
            self.assertFalse(response is redirect)
            self.assertContains(response, "302 Found")
            self.assertContains(response, "http://somewhere/else/")

    def test_unknown_status_code(self):
        redirect = HttpResponse(status=369)
        redirect["Location"] = "http://somewhere/else/"
        self._get_response = lambda request: redirect
        response = self.panel.process_request(self.request)
        self.assertContains(response, "369 Unknown Status Code")

    def test_unknown_status_code_with_reason(self):
        redirect = HttpResponse(status=369, reason="Look Ma!")
        redirect["Location"] = "http://somewhere/else/"
        self._get_response = lambda request: redirect
        response = self.panel.process_request(self.request)
        self.assertContains(response, "369 Look Ma!")

    async def test_async_compatibility(self):
        redirect = HttpResponse(status=302)

        async def get_response(request):
            return redirect

        await_response = await get_response(self.request)
        self._get_response = get_response

        self.request = AsyncRequestFactory().get("/")
        response = await self.panel.process_request(self.request)
        self.assertIsInstance(response, HttpResponse)
        self.assertTrue(response is await_response)

    def test_original_response_preserved(self):
        redirect = HttpResponse(status=302)
        redirect["Location"] = "http://somewhere/else/"
        self._get_response = lambda request: redirect
        response = self.panel.process_request(self.request)
        self.assertFalse(response is redirect)
        self.assertTrue(hasattr(response, "original_response"))
        self.assertTrue(response.original_response is redirect)
        self.assertIsNone(response.get("Location"))
        self.assertEqual(
            response.original_response.get("Location"), "http://somewhere/else/"
        )

    def test_deprecation_warning(self):
        """Test that a deprecation warning is shown when RedirectsPanel used."""
        redirect = HttpResponse(status=304)
        redirect["Location"] = "http://somewhere/else/"
        self._get_response = lambda request: redirect

        with self.assertWarns(DeprecationWarning) as cm:
            self.panel.process_request(self.request)

        self.assertIn("RedirectsPanel is deprecated", str(cm.warning))
        self.assertIn("HistoryPanel", str(cm.warning))



from django.contrib.auth.models import User
from django.http import JsonResponse


async def async_db_view(request):
    names = []
    async for user in User.objects.all():
        names.append(user.username)
    return JsonResponse({"names": names})



from unittest.mock import patch

from django.http import Http404, HttpResponse
from django.test import AsyncRequestFactory, RequestFactory, TestCase
from django.test.utils import override_settings

from debug_toolbar.decorators import render_with_toolbar_language, require_show_toolbar


@render_with_toolbar_language
def stub_view(request):
    return HttpResponse(200)


@require_show_toolbar
def stub_require_toolbar_view(request):
    return HttpResponse(200)


@require_show_toolbar
async def stub_require_toolbar_async_view(request):
    return HttpResponse(200)


class TestRequireToolbar(TestCase):
    """
    Tests require_toolbar functionality and async compatibility.
    """

    def setUp(self):
        self.factory = RequestFactory()
        self.async_factory = AsyncRequestFactory()

    @override_settings(DEBUG=True)
    def test_require_toolbar_debug_true(self):
        response = stub_require_toolbar_view(self.factory.get("/"))
        self.assertEqual(response.status_code, 200)

    def test_require_toolbar_debug_false(self):
        with self.assertRaises(Http404):
            stub_require_toolbar_view(self.factory.get("/"))

    # Following tests additionally tests async compatibility
    # of require_toolbar decorator
    @override_settings(DEBUG=True)
    async def test_require_toolbar_async_debug_true(self):
        response = await stub_require_toolbar_async_view(self.async_factory.get("/"))
        self.assertEqual(response.status_code, 200)

    async def test_require_toolbar_async_debug_false(self):
        with self.assertRaises(Http404):
            await stub_require_toolbar_async_view(self.async_factory.get("/"))


@override_settings(DEBUG=True, LANGUAGE_CODE="fr")
class RenderWithToolbarLanguageTestCase(TestCase):
    @override_settings(DEBUG_TOOLBAR_CONFIG={"TOOLBAR_LANGUAGE": "de"})
    @patch("debug_toolbar.decorators.language_override")
    def test_uses_toolbar_language(self, mock_language_override):
        stub_view(RequestFactory().get("/"))
        mock_language_override.assert_called_once_with("de")

    @patch("debug_toolbar.decorators.language_override")
    def test_defaults_to_django_language_code(self, mock_language_override):
        stub_view(RequestFactory().get("/"))
        mock_language_override.assert_called_once_with("fr")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import datetime
import os
import sys

os.environ["DJANGO_SETTINGS_MODULE"] = "example.settings"
sys.path.append(os.path.dirname(os.path.dirname(__file__)))


# -- Project information -----------------------------------------------------

project = "Django Debug Toolbar"
copyright = "{}, Django Debug Toolbar developers and contributors"
copyright = copyright.format(datetime.datetime.now(datetime.timezone.utc).year)

# The full version, including alpha/beta/rc tags
release = "6.2.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# Holds the brand assets (logos) referenced by the design guidelines.
html_static_path = ["_static"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "django": (
        "https://docs.djangoproject.com/en/dev/",
        "https://docs.djangoproject.com/en/dev/_objects/",
    ),
}

# -- Options for Read the Docs -----------------------------------------------


def setup(app):
    """Configure documentation via Sphinx extension"""
    # Add the :setting: role for intersphinx linking to Django's docs
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )



import os
import sys
import warnings
from functools import cache

from django.conf import settings
from django.dispatch import receiver
from django.test.signals import setting_changed


def _is_running_tests():
    """
    Helper function to support testing default value for
    IS_RUNNING_TESTS
    """
    return "test" in sys.argv or "PYTEST_VERSION" in os.environ


CONFIG_DEFAULTS = {
    # Toolbar options
    "CACHE_BACKEND": "default",
    "CACHE_KEY_PREFIX": "djdt:",
    "DISABLE_PANELS": {
        "debug_toolbar.panels.profiling.ProfilingPanel",
        "debug_toolbar.panels.redirects.RedirectsPanel",
    },
    "INSERT_BEFORE": "</body>",
    "IS_RUNNING_TESTS": _is_running_tests(),
    "OBSERVE_REQUEST_CALLBACK": "debug_toolbar.toolbar.observe_request",
    "RENDER_PANELS": None,
    "RESULTS_CACHE_SIZE": 25,
    "ROOT_TAG_EXTRA_ATTRS": "",
    "SHOW_COLLAPSED": False,
    "SHOW_TOOLBAR_CALLBACK": "debug_toolbar.middleware.show_toolbar",
    "USE_SHADOW_DOM": True,
    "TOOLBAR_LANGUAGE": None,
    "TOOLBAR_STORE_CLASS": "debug_toolbar.store.MemoryStore",
    "UPDATE_ON_FETCH": False,
    # Panel options
    "EXTRA_SIGNALS": [],
    "ENABLE_STACKTRACES": True,
    "ENABLE_STACKTRACES_LOCALS": False,
    "HIDE_IN_STACKTRACES": (
        "socketserver",
        "threading",
        "wsgiref",
        "debug_toolbar",
        "django.db",
        "django.core.handlers",
        "django.core.servers",
        "django.utils.decorators",
        "django.utils.deprecation",
        "django.utils.functional",
    ),
    "PRETTIFY_SQL": True,
    "PROFILER_CAPTURE_PROJECT_CODE": True,
    "PROFILER_MAX_DEPTH": 10,
    "PROFILER_THRESHOLD_RATIO": 8,
    "SHOW_TEMPLATE_CONTEXT": True,
    "SKIP_TEMPLATE_PREFIXES": ("django/forms/widgets/", "admin/widgets/"),
    "SKIP_TOOLBAR_QUERIES": True,
    "SQL_WARNING_THRESHOLD": 500,  # milliseconds
}


@cache
def get_config():
    USER_CONFIG = getattr(settings, "DEBUG_TOOLBAR_CONFIG", {})
    CONFIG = CONFIG_DEFAULTS.copy()
    CONFIG.update(USER_CONFIG)
    return CONFIG


PANELS_DEFAULTS = [
    "debug_toolbar.panels.history.HistoryPanel",
    "debug_toolbar.panels.versions.VersionsPanel",
    "debug_toolbar.panels.timer.TimerPanel",
    "debug_toolbar.panels.settings.SettingsPanel",
    "debug_toolbar.panels.headers.HeadersPanel",
    "debug_toolbar.panels.request.RequestPanel",
    "debug_toolbar.panels.sql.SQLPanel",
    "debug_toolbar.panels.staticfiles.StaticFilesPanel",
    "debug_toolbar.panels.templates.TemplatesPanel",
    "debug_toolbar.panels.alerts.AlertsPanel",
    "debug_toolbar.panels.cache.CachePanel",
    "debug_toolbar.panels.signals.SignalsPanel",
    "debug_toolbar.panels.tasks.TasksPanel",
    "debug_toolbar.panels.community.CommunityPanel",
    "debug_toolbar.panels.redirects.RedirectsPanel",
    "debug_toolbar.panels.profiling.ProfilingPanel",
]


@cache
def get_panels():
    try:
        PANELS = list(settings.DEBUG_TOOLBAR_PANELS)
    except AttributeError:
        PANELS = PANELS_DEFAULTS

    logging_panel = "debug_toolbar.panels.logging.LoggingPanel"
    if logging_panel in PANELS:
        PANELS = [panel for panel in PANELS if panel != logging_panel]
        warnings.warn(
            f"Please remove {logging_panel} from your DEBUG_TOOLBAR_PANELS setting.",
            DeprecationWarning,
            stacklevel=1,
        )
    return PANELS


@receiver(setting_changed)
def update_toolbar_config(*, setting, **kwargs):
    """
    Refresh configuration when overriding settings.
    """
    if setting == "DEBUG_TOOLBAR_CONFIG":
        get_config.cache_clear()
    elif setting == "DEBUG_TOOLBAR_PANELS":
        from debug_toolbar.toolbar import DebugToolbar

        get_panels.cache_clear()
        DebugToolbar._panel_classes = None
        # Not implemented: invalidate debug_toolbar.urls.



from django import forms


class HistoryStoreForm(forms.Form):
    """
    Validate params

        request_id: The key for the store instance to be fetched.
    """

    request_id = forms.CharField(widget=forms.HiddenInput())
    exclude_history = forms.BooleanField(widget=forms.HiddenInput(), required=False)



from django.contrib.auth.models import User
from django.template.loaders.app_directories import Loader


class LoaderWithSQL(Loader):
    def get_template(self, *args, **kwargs):
        # Force the template loader to run some SQL. Simulates a CMS.
        User.objects.all().count()
        return super().get_template(*args, **kwargs)



import asyncio
import datetime
import os
import unittest
from unittest.mock import call, patch

import django
from asgiref.sync import sync_to_async
from django.apps import apps
from django.contrib.auth.models import User
from django.db import connection, transaction
from django.db.backends.utils import CursorDebugWrapper, CursorWrapper
from django.db.models import Count
from django.db.utils import DatabaseError
from django.shortcuts import render
from django.test.utils import override_settings
from sqlparse.exceptions import SQLParseError

import debug_toolbar.panels.sql.tracking as sql_tracking
from debug_toolbar import settings as dt_settings
from debug_toolbar.models import HistoryEntry
from debug_toolbar.panels.sql import SQLPanel, tracking
from debug_toolbar.panels.sql.utils import parse_sql

try:
    import psycopg
except ImportError:
    psycopg = None

from ..base import BaseMultiDBTestCase, BaseTestCase
from ..models import Binary, PostgresJSON


def sql_call(*, use_iterator=False):
    qs = User.objects.all()
    if use_iterator:
        qs = qs.iterator()
    return list(qs)


def sql_call_toolbar_model():
    """Query one of the toolbar's models to test tracking of SQL queries."""
    qs = HistoryEntry.objects.all()
    return list(qs)


async def async_sql_call_toolbar_model():
    """(async) Query one of the toolbar's models to test tracking of SQL queries."""
    qs = HistoryEntry.objects.all()
    return await sync_to_async(list)(qs)


async def async_sql_call(*, use_iterator=False):
    qs = User.objects.all()
    if use_iterator:
        qs = qs.iterator()
    return await sync_to_async(list)(qs)


async def concurrent_async_sql_call(*, use_iterator=False):
    qs = User.objects.all()
    if use_iterator:
        qs = qs.iterator()
    return await asyncio.gather(sync_to_async(list)(qs), User.objects.acount())


def patch_tracking_ddt_models():
    """
    Set the tracking.DDT_MODELS constant to the toolbar's models.
    This only gets configured when the store is set to the DatabaseStore.
    """
    if (
        dt_settings.get_config()["TOOLBAR_STORE_CLASS"]
        == "debug_toolbar.store.DatabaseStore"
    ):
        apps.get_app_config("debug_toolbar").import_models()
        tracking.DDT_MODELS = {
            m._meta.db_table for m in apps.get_app_config("debug_toolbar").get_models()
        }


class SQLPanelTestCase(BaseTestCase):
    panel_id = SQLPanel.panel_id

    def test_disabled(self):
        config = {"DISABLE_PANELS": {"debug_toolbar.panels.sql.SQLPanel"}}
        self.assertTrue(self.panel.enabled)
        with self.settings(DEBUG_TOOLBAR_CONFIG=config):
            self.assertFalse(self.panel.enabled)

    def test_recording(self):
        self.assertEqual(len(self.panel._queries), 0)

        sql_call()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertTrue("duration" in query)
        self.assertTrue("stacktrace" in query)

        # ensure the stacktrace is populated
        self.assertTrue(len(query["stacktrace"]) > 0)

    def test_assert_num_queries_works(self):
        """
        Confirm Django's assertNumQueries and CaptureQueriesContext works

        See  https://github.com/django-commons/django-debug-toolbar/issues/1791
        """
        self.assertEqual(len(self.panel._queries), 0)
        with self.assertNumQueries(1):
            sql_call()
        self.assertEqual(len(self.panel._queries), 1)

    async def test_recording_async(self):
        self.assertEqual(len(self.panel._queries), 0)

        await async_sql_call()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertTrue("duration" in query)
        self.assertTrue("stacktrace" in query)

        # ensure the stacktrace is populated
        self.assertTrue(len(query["stacktrace"]) > 0)

    async def test_recording_concurrent_async(self):
        self.assertEqual(len(self.panel._queries), 0)

        await concurrent_async_sql_call()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 2)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertTrue("duration" in query)
        self.assertTrue("stacktrace" in query)

        # ensure the stacktrace is populated
        self.assertTrue(len(query["stacktrace"]) > 0)

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "SKIP_TOOLBAR_QUERIES": False,
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore",
        }
    )
    def test_toolbar_model_query_is_tracked(self):
        """
        Test is toolbar models are tracked when the `SKIP_TOOLBAR_QUERIES`
        is set to False.
        """
        self.assertEqual(len(self.panel._queries), 0)

        patch_tracking_ddt_models()
        sql_call_toolbar_model()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertTrue(HistoryEntry._meta.db_table in query["sql"])

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "SKIP_TOOLBAR_QUERIES": False,
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore",
        }
    )
    async def test_toolbar_model_query_is_tracked_async(self):
        """
        (async) Test is toolbar models are tracked when the `SKIP_TOOLBAR_QUERIES`
        is set to False.
        """
        self.assertEqual(len(self.panel._queries), 0)

        patch_tracking_ddt_models()
        await async_sql_call_toolbar_model()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertTrue(HistoryEntry._meta.db_table in query["sql"])

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "SKIP_TOOLBAR_QUERIES": True,
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore",
        }
    )
    def test_toolbar_model_query_is_not_tracked(self):
        """
        Test is toolbar models are not tracked when the `SKIP_TOOLBAR_QUERIES`
        is set to True.
        """
        self.assertEqual(len(self.panel._queries), 0)

        patch_tracking_ddt_models()
        sql_call_toolbar_model()

        self.assertEqual(len(self.panel._queries), 0)

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "SKIP_TOOLBAR_QUERIES": True,
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore",
        }
    )
    async def test_toolbar_model_query_is_not_tracked_async(self):
        """
        (async) Test is toolbar models are not tracked when the `SKIP_TOOLBAR_QUERIES`
        is set to True.
        """
        self.assertEqual(len(self.panel._queries), 0)

        patch_tracking_ddt_models()
        await async_sql_call_toolbar_model()

        self.assertEqual(len(self.panel._queries), 0)

    @unittest.skipUnless(
        connection.vendor == "postgresql", "Test valid only on PostgreSQL"
    )
    def test_recording_chunked_cursor(self):
        self.assertEqual(len(self.panel._queries), 0)

        sql_call(use_iterator=True)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)

    @patch(
        "debug_toolbar.panels.sql.tracking.patch_cursor_wrapper_with_mixin",
        wraps=sql_tracking.patch_cursor_wrapper_with_mixin,
    )
    def test_cursor_wrapper_singleton(self, mock_patch_cursor_wrapper):
        sql_call()
        # ensure that cursor wrapping is applied only once
        self.assertIn(
            mock_patch_cursor_wrapper.mock_calls,
            [
                [call(CursorWrapper, sql_tracking.NormalCursorMixin)],
                # CursorDebugWrapper is used if the test is called with `--debug-sql`
                [call(CursorDebugWrapper, sql_tracking.NormalCursorMixin)],
            ],
        )

    @patch(
        "debug_toolbar.panels.sql.tracking.patch_cursor_wrapper_with_mixin",
        wraps=sql_tracking.patch_cursor_wrapper_with_mixin,
    )
    def test_chunked_cursor_wrapper_singleton(self, mock_patch_cursor_wrapper):
        sql_call(use_iterator=True)

        # ensure that cursor wrapping is applied only once
        self.assertIn(
            mock_patch_cursor_wrapper.mock_calls,
            [
                [call(CursorWrapper, sql_tracking.NormalCursorMixin)],
                # CursorDebugWrapper is used if the test is called with `--debug-sql`
                [call(CursorDebugWrapper, sql_tracking.NormalCursorMixin)],
            ],
        )

    @patch(
        "debug_toolbar.panels.sql.tracking.patch_cursor_wrapper_with_mixin",
        wraps=sql_tracking.patch_cursor_wrapper_with_mixin,
    )
    async def test_cursor_wrapper_async(self, mock_patch_cursor_wrapper):
        await sync_to_async(sql_call)()

        self.assertIn(
            mock_patch_cursor_wrapper.mock_calls,
            [
                [call(CursorWrapper, sql_tracking.NormalCursorMixin)],
                # CursorDebugWrapper is used if the test is called with `--debug-sql`
                [call(CursorDebugWrapper, sql_tracking.NormalCursorMixin)],
            ],
        )

    @patch(
        "debug_toolbar.panels.sql.tracking.patch_cursor_wrapper_with_mixin",
        wraps=sql_tracking.patch_cursor_wrapper_with_mixin,
    )
    async def test_cursor_wrapper_asyncio_ctx(self, mock_patch_cursor_wrapper):
        self.assertTrue(sql_tracking.allow_sql.get())
        await sync_to_async(sql_call)()

        async def task():
            sql_tracking.allow_sql.set(False)
            # By disabling sql_tracking.allow_sql, we are indicating that any
            # future SQL queries should be stopped. If SQL query occurs,
            # it raises an exception.
            with self.assertRaises(sql_tracking.SQLQueryTriggered):
                await sync_to_async(sql_call)()

        # Ensure this is called in another context
        await asyncio.create_task(task())
        # Because it was called in another context, it should not have affected ours
        self.assertTrue(sql_tracking.allow_sql.get())

        self.assertIn(
            mock_patch_cursor_wrapper.mock_calls,
            [
                [
                    call(CursorWrapper, sql_tracking.NormalCursorMixin),
                    call(CursorWrapper, sql_tracking.ExceptionCursorMixin),
                ],
                # CursorDebugWrapper is used if the test is called with `--debug-sql`
                [
                    call(CursorDebugWrapper, sql_tracking.NormalCursorMixin),
                    call(CursorDebugWrapper, sql_tracking.ExceptionCursorMixin),
                ],
            ],
        )

    def test_generate_server_timing(self):
        self.assertEqual(len(self.panel._queries), 0)

        sql_call()

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.panel.generate_server_timing(self.request, response)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]

        expected_data = {
            "sql_time": {"title": "SQL 1 queries", "value": query["duration"]}
        }

        self.assertEqual(self.panel.get_server_timing_stats(), expected_data)

    def test_non_ascii_query(self):
        self.assertEqual(len(self.panel._queries), 0)

        # non-ASCII text query
        list(User.objects.extra(where=["username = 'apéro'"]))
        self.assertEqual(len(self.panel._queries), 1)

        # non-ASCII text parameters
        list(User.objects.filter(username="thé"))
        self.assertEqual(len(self.panel._queries), 2)

        # non-ASCII bytes parameters
        list(Binary.objects.filter(field__in=["café".encode()]))
        self.assertEqual(len(self.panel._queries), 3)

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # ensure the panel renders correctly
        self.assertIn("café", self.panel.content)

    @unittest.skipUnless(
        connection.vendor == "postgresql", "Test valid only on PostgreSQL"
    )
    def test_bytes_query(self):
        self.assertEqual(len(self.panel._queries), 0)

        with connection.cursor() as cursor:
            cursor.execute(b"SELECT 1")

        self.assertEqual(len(self.panel._queries), 1)

    def test_param_conversion(self):
        self.assertEqual(len(self.panel._queries), 0)

        list(
            User.objects.filter(first_name="Foo")
            .filter(is_staff=True)
            .filter(is_superuser=False)
        )
        list(
            User.objects.annotate(group_count=Count("groups__id"))
            .filter(group_count__lt=10)
            .filter(group_count__gt=1)
        )
        list(
            User.objects.filter(
                date_joined=datetime.datetime(
                    2017, 12, 22, 16, 7, 1, tzinfo=datetime.timezone.utc
                )
            )
        )

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 3)

        if (
            connection.vendor == "mysql"
            and django.VERSION >= (4, 1)
            or connection.vendor != "postgresql"
            and django.VERSION >= (6, 1)
        ):
            # Django 4.1 started passing true/false back for boolean
            # comparisons in MySQL.
            # Django 6.1 started passing true/false for all-non
            # postgres databases.
            expected_bools = ["Foo", True, False]
        else:
            expected_bools = ["Foo"]

        if connection.vendor == "postgresql":
            # PostgreSQL always includes timezone
            expected_datetime = ["2017-12-22 16:07:01+00:00"]
        else:
            expected_datetime = ["2017-12-22 16:07:01"]

        self.assertEqual(
            tuple(query["params"] for query in self.panel._queries),
            (
                expected_bools,
                [10, 1],
                expected_datetime,
            ),
        )

    @unittest.skipUnless(
        connection.vendor == "postgresql", "Test valid only on PostgreSQL"
    )
    def test_json_param_conversion(self):
        self.assertEqual(len(self.panel._queries), 0)

        list(PostgresJSON.objects.filter(field__contains={"foo": "bar"}))

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        self.assertEqual(
            self.panel._queries[0]["params"],
            ['{"foo": "bar"}'],
        )

    @unittest.skipUnless(
        connection.vendor == "postgresql" and psycopg is None,
        "Test valid only on PostgreSQL with psycopg2",
    )
    def test_tuple_param_conversion(self):
        """
        Regression test for tuple parameter conversion.
        """
        self.assertEqual(len(self.panel._queries), 0)

        list(
            PostgresJSON.objects.raw(
                "SELECT * FROM tests_postgresjson WHERE field ->> 'key' IN %s",
                [("a", "b'")],
            )
        )

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        self.assertEqual(self.panel._queries[0]["params"], [["a", "b'"]])

    def test_binary_param_force_text(self):
        self.assertEqual(len(self.panel._queries), 0)

        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT * FROM tests_binary WHERE field = %s",
                [connection.Database.Binary(b"\xff")],
            )

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        self.assertEqual(len(self.panel._queries), 1)
        self.assertIn(
            "<strong>SELECT</strong> * <strong>FROM</strong>"
            " tests_binary <strong>WHERE</strong> field =",
            self.panel.content,
        )

    @unittest.skipUnless(connection.vendor != "sqlite", "Test invalid for SQLite")
    def test_raw_query_param_conversion(self):
        self.assertEqual(len(self.panel._queries), 0)

        list(
            User.objects.raw(
                "SELECT * FROM auth_user WHERE first_name = %s AND is_staff = %s AND is_superuser = %s AND date_joined = %s",
                params=[
                    "Foo",
                    True,
                    False,
                    datetime.datetime(2017, 12, 22, 16, 7, 1),  # noqa: DTZ001
                ],
            )
        )

        list(
            User.objects.raw(
                "SELECT * FROM auth_user WHERE first_name = %(first_name)s AND is_staff = %(is_staff)s AND is_superuser = %(is_superuser)s AND date_joined = %(date_joined)s",
                params={
                    "first_name": "Foo",
                    "is_staff": True,
                    "is_superuser": False,
                    "date_joined": datetime.datetime(2017, 12, 22, 16, 7, 1),  # noqa: DTZ001
                },
            )
        )

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 2)

        self.assertEqual(
            tuple(query["params"] for query in self.panel._queries),
            (
                ["Foo", True, False, "2017-12-22 16:07:01"],
                {
                    "first_name": "Foo",
                    "is_staff": True,
                    "is_superuser": False,
                    "date_joined": "2017-12-22 16:07:01",
                },
            ),
        )

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        list(User.objects.filter(username="café"))
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn("café", content)
        self.assertValidHTML(content)

    @override_settings(DEBUG_TOOLBAR_CONFIG={"ENABLE_STACKTRACES_LOCALS": True})
    def test_insert_locals(self):
        """
        Test that the panel inserts locals() content.
        """
        local_var = "<script>alert('test');</script>"  # noqa: F841
        list(User.objects.filter(username="café"))
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIn("local_var", self.panel.content)
        # Verify the escape logic works
        content = self.panel.content
        self.assertNotIn("<script>alert", content)
        self.assertIn("&lt;script&gt;alert", content)
        self.assertIn("djdt-locals", content)

    def test_not_insert_locals(self):
        """
        Test that the panel does not insert locals() content.
        """
        list(User.objects.filter(username="café"))
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertNotIn("djdt-locals", self.panel.content)

    @unittest.skipUnless(
        connection.vendor == "postgresql", "Test valid only on PostgreSQL"
    )
    def test_erroneous_query(self):
        """
        Test that an error in the query isn't swallowed by the middleware.
        """
        try:
            connection.cursor().execute("erroneous query")
        except DatabaseError as e:
            self.assertTrue("erroneous query" in str(e))

    @unittest.skipUnless(
        connection.vendor == "postgresql", "Test valid only on PostgreSQL"
    )
    def test_execute_with_psycopg_composed_sql(self):
        """
        Test command executed using a Composed psycopg object is logged.
        Ref: https://www.psycopg.org/psycopg3/docs/api/sql.html
        """
        try:
            from psycopg import sql
        except ImportError:
            from psycopg2 import sql

        self.assertEqual(len(self.panel._queries), 0)

        with connection.cursor() as cursor:
            command = sql.SQL("select {field} from {table}").format(
                field=sql.Identifier("username"), table=sql.Identifier("auth_user")
            )
            cursor.execute(command)

        self.assertEqual(len(self.panel._queries), 1)

        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertEqual(query["sql"], 'select "username" from "auth_user"')

    def test_disable_stacktraces(self):
        self.assertEqual(len(self.panel._queries), 0)

        with self.settings(DEBUG_TOOLBAR_CONFIG={"ENABLE_STACKTRACES": False}):
            sql_call()

        # ensure query was logged
        self.assertEqual(len(self.panel._queries), 1)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertTrue("duration" in query)
        self.assertTrue("stacktrace" in query)

        # ensure the stacktrace is empty
        self.assertEqual([], query["stacktrace"])

    @override_settings(
        DEBUG=True,
        TEMPLATES=[
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "OPTIONS": {"debug": True, "loaders": ["tests.loaders.LoaderWithSQL"]},
            }
        ],
    )
    def test_regression_infinite_recursion(self):
        """
        Test case for when the template loader runs a SQL query that causes
        an infinite recursion in the SQL panel.
        """
        self.assertEqual(len(self.panel._queries), 0)

        render(self.request, "basic.html", {})

        # Two queries are logged because the loader runs SQL every time a
        # template is loaded and basic.html extends base.html.
        self.assertEqual(len(self.panel._queries), 2)
        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")
        self.assertTrue("sql" in query)
        self.assertTrue("duration" in query)
        self.assertTrue("stacktrace" in query)

        # ensure the stacktrace is populated
        self.assertTrue(len(query["stacktrace"]) > 0)

    def test_prettify_sql(self):
        """
        Test case to validate that the PRETTIFY_SQL setting changes the output
        of the sql when it's toggled. It does not validate what it does
        though.
        """
        with override_settings(DEBUG_TOOLBAR_CONFIG={"PRETTIFY_SQL": True}):
            list(User.objects.filter(username__istartswith="spam"))
            response = self.panel.process_request(self.request)
            self.panel.generate_stats(self.request, response)
            # The content formats the sql and prettifies it
            self.assertTrue(self.panel.content)
            pretty_sql = self.panel._queries[-1]["sql"]
            self.assertEqual(len(self.panel._queries), 1)

        # Recreate the panel to reset the queries. Content being a cached_property
        # which doesn't have a way to reset it.
        self.panel.disable_instrumentation()
        self.panel = SQLPanel(self.panel.toolbar, self.panel.get_response)
        self.panel.enable_instrumentation()
        # Run it again, but with prettify off. Verify that it's different.
        with override_settings(DEBUG_TOOLBAR_CONFIG={"PRETTIFY_SQL": False}):
            list(User.objects.filter(username__istartswith="spam"))
            response = self.panel.process_request(self.request)
            self.panel.generate_stats(self.request, response)
            # The content formats the sql and prettifies it
            self.assertTrue(self.panel.content)
            self.assertEqual(len(self.panel._queries), 1)
            self.assertNotIn(pretty_sql, self.panel.content)

        self.panel.disable_instrumentation()
        self.panel = SQLPanel(self.panel.toolbar, self.panel.get_response)
        self.panel.enable_instrumentation()
        # Run it again, but with prettify back on.
        # This is so we don't have to check what PRETTIFY_SQL does exactly,
        # but we know it's doing something.
        with override_settings(DEBUG_TOOLBAR_CONFIG={"PRETTIFY_SQL": True}):
            list(User.objects.filter(username__istartswith="spam"))
            response = self.panel.process_request(self.request)
            self.panel.generate_stats(self.request, response)
            # The content formats the sql and prettifies it
            self.assertTrue(self.panel.content)
            self.assertEqual(len(self.panel._queries), 1)
            self.assertIn(pretty_sql, self.panel.content)

    def test_simplification(self):
        """
        Test case to validate that select lists for .count() and .exist() queries do not
        get elided, but other select lists do.
        """
        User.objects.count()
        User.objects.exists()
        list(User.objects.values_list("id"))
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # The content formats the sql which injects the ellipsis character
        self.assertTrue(self.panel.content)
        self.assertEqual(len(self.panel._queries), 3)
        self.assertNotIn("\u2022", self.panel._queries[0]["sql"])
        self.assertNotIn("\u2022", self.panel._queries[1]["sql"])
        self.assertIn("\u2022", self.panel._queries[2]["sql"])

    def test_top_level_simplification(self):
        """
        Test case to validate that top-level select lists get elided, but other select
        lists for subselects do not.
        """
        list(User.objects.filter(id__in=User.objects.filter(is_staff=True)))
        list(User.objects.filter(id__lt=20).union(User.objects.filter(id__gt=10)))
        if connection.vendor != "mysql":
            list(
                User.objects.filter(id__lt=20).intersection(
                    User.objects.filter(id__gt=10)
                )
            )
            list(
                User.objects.filter(id__lt=20).difference(
                    User.objects.filter(id__gt=10)
                )
            )
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # The content formats the sql which injects the ellipsis character
        self.assertTrue(self.panel.content)
        if connection.vendor != "mysql":
            self.assertEqual(len(self.panel._queries), 4)
        else:
            self.assertEqual(len(self.panel._queries), 2)
        # WHERE ... IN SELECT ... queries should have only one elided select list
        self.assertEqual(self.panel._queries[0]["sql"].count("SELECT"), 4)
        self.assertEqual(self.panel._queries[0]["sql"].count("\u2022"), 3)
        # UNION queries should have two elidid select lists
        self.assertEqual(self.panel._queries[1]["sql"].count("SELECT"), 4)
        self.assertEqual(self.panel._queries[1]["sql"].count("\u2022"), 6)
        if connection.vendor != "mysql":
            # INTERSECT queries should have two elidid select lists
            self.assertEqual(self.panel._queries[2]["sql"].count("SELECT"), 4)
            self.assertEqual(self.panel._queries[2]["sql"].count("\u2022"), 6)
            # EXCEPT queries should have two elidid select lists
            self.assertEqual(self.panel._queries[3]["sql"].count("SELECT"), 4)
            self.assertEqual(self.panel._queries[3]["sql"].count("\u2022"), 6)

    @override_settings(
        DEBUG=True,
    )
    def test_flat_template_information(self):
        """
        Test case for when the query is used in a flat template hierarchy
        (without included templates).
        """
        self.assertEqual(len(self.panel._queries), 0)

        users = User.objects.all()
        render(self.request, "sql/flat.html", {"users": users})

        self.assertEqual(len(self.panel._queries), 1)

        query = self.panel._queries[0]
        template_info = query["template_info"]
        template_name = os.path.basename(template_info["name"])
        self.assertEqual(template_name, "flat.html")
        self.assertEqual(template_info["context"][3]["content"].strip(), "{{ users }}")
        self.assertEqual(template_info["context"][3]["highlight"], True)

    @override_settings(
        DEBUG=True,
    )
    def test_nested_template_information(self):
        """
        Test case for when the query is used in a nested template
        hierarchy (with included templates).
        """
        self.assertEqual(len(self.panel._queries), 0)

        users = User.objects.all()
        render(self.request, "sql/nested.html", {"users": users})

        self.assertEqual(len(self.panel._queries), 1)

        query = self.panel._queries[0]
        template_info = query["template_info"]
        template_name = os.path.basename(template_info["name"])
        self.assertEqual(template_name, "included.html")
        self.assertEqual(template_info["context"][0]["content"].strip(), "{{ users }}")
        self.assertEqual(template_info["context"][0]["highlight"], True)

    def test_similar_and_duplicate_grouping(self):
        self.assertEqual(len(self.panel._queries), 0)

        User.objects.filter(id=1).count()
        User.objects.filter(id=1).count()
        User.objects.filter(id=2).count()
        User.objects.filter(id__lt=10).count()
        User.objects.filter(id__lt=20).count()
        User.objects.filter(id__gt=10, id__lt=20).count()

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        self.assertEqual(len(self.panel._queries), 6)

        queries = self.panel._queries
        query = queries[0]
        self.assertEqual(query["similar_count"], 3)
        self.assertEqual(query["duplicate_count"], 2)

        query = queries[1]
        self.assertEqual(query["similar_count"], 3)
        self.assertEqual(query["duplicate_count"], 2)

        query = queries[2]
        self.assertEqual(query["similar_count"], 3)
        self.assertTrue("duplicate_count" not in query)

        query = queries[3]
        self.assertEqual(query["similar_count"], 2)
        self.assertTrue("duplicate_count" not in query)

        query = queries[4]
        self.assertEqual(query["similar_count"], 2)
        self.assertTrue("duplicate_count" not in query)

        query = queries[5]
        self.assertTrue("similar_count" not in query)
        self.assertTrue("duplicate_count" not in query)

        self.assertEqual(queries[0]["similar_color"], queries[1]["similar_color"])
        self.assertEqual(queries[0]["similar_color"], queries[2]["similar_color"])
        self.assertEqual(queries[0]["duplicate_color"], queries[1]["duplicate_color"])
        self.assertNotEqual(queries[0]["similar_color"], queries[0]["duplicate_color"])

        self.assertEqual(queries[3]["similar_color"], queries[4]["similar_color"])
        self.assertNotEqual(queries[0]["similar_color"], queries[3]["similar_color"])
        self.assertNotEqual(queries[0]["duplicate_color"], queries[3]["similar_color"])

    def test_explain_with_union(self):
        list(User.objects.filter(id__lt=20).union(User.objects.filter(id__gt=10)))
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIn("Expl", self.panel.content)

    @override_settings(DEBUG_TOOLBAR_CONFIG={"PRETTIFY_SQL": True})
    def test_sql_parse_error_graceful_degradation(self):
        """
        Test that SQLParseError is handled gracefully by disabling grouping.
        """
        parse_sql.cache_clear()

        def run_side_effect(sql):
            if mock_stack.run.call_count == 1:
                raise SQLParseError("Token limit exceeded")
            return [sql]

        with patch("debug_toolbar.panels.sql.utils.get_filter_stack") as mock_get_stack:
            mock_stack = mock_get_stack.return_value
            mock_stack.run.side_effect = run_side_effect
            mock_stack._grouping = True

            result = parse_sql("SELECT * FROM test")

            # Should have been called twice (once for error, once for retry)
            self.assertEqual(mock_stack.run.call_count, 2)
            # On retry, _grouping should be set to False
            self.assertFalse(mock_stack._grouping)
            self.assertIn("SELECT", result)

        parse_sql.cache_clear()


class SQLPanelMultiDBTestCase(BaseMultiDBTestCase):
    panel_id = SQLPanel.panel_id

    def test_aliases(self):
        self.assertFalse(self.panel._queries)

        list(User.objects.all())
        list(User.objects.using("replica").all())

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        self.assertTrue(self.panel._queries)

        query = self.panel._queries[0]
        self.assertEqual(query["alias"], "default")

        query = self.panel._queries[-1]
        self.assertEqual(query["alias"], "replica")

    def test_transaction_status(self):
        """
        Test case for tracking the transaction status is properly associated with
        queries on PostgreSQL, and that transactions aren't broken on other database
        engines.
        """
        self.assertEqual(len(self.panel._queries), 0)

        with transaction.atomic():
            list(User.objects.all())
            list(User.objects.using("replica").all())

            with transaction.atomic(using="replica"):
                list(User.objects.all())
                list(User.objects.using("replica").all())

        with transaction.atomic():
            list(User.objects.all())

        list(User.objects.using("replica").all())

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        if connection.vendor == "postgresql":
            # Connection tracking is currently only implemented for PostgreSQL.
            self.assertEqual(len(self.panel._queries), 6)

            query = self.panel._queries[0]
            self.assertEqual(query["alias"], "default")
            self.assertIsNotNone(query["trans_id"])
            self.assertTrue(query["starts_trans"])
            self.assertTrue(query["in_trans"])
            self.assertFalse("end_trans" in query)

            query = self.panel._queries[-1]
            self.assertEqual(query["alias"], "replica")
            self.assertIsNone(query["trans_id"])
            self.assertFalse("starts_trans" in query)
            self.assertFalse("in_trans" in query)
            self.assertFalse("end_trans" in query)

            query = self.panel._queries[2]
            self.assertEqual(query["alias"], "default")
            self.assertIsNotNone(query["trans_id"])
            self.assertEqual(query["trans_id"], self.panel._queries[0]["trans_id"])
            self.assertFalse("starts_trans" in query)
            self.assertTrue(query["in_trans"])
            self.assertTrue(query["ends_trans"])

            query = self.panel._queries[3]
            self.assertEqual(query["alias"], "replica")
            self.assertIsNotNone(query["trans_id"])
            self.assertNotEqual(query["trans_id"], self.panel._queries[0]["trans_id"])
            self.assertTrue(query["starts_trans"])
            self.assertTrue(query["in_trans"])
            self.assertTrue(query["ends_trans"])

            query = self.panel._queries[4]
            self.assertEqual(query["alias"], "default")
            self.assertIsNotNone(query["trans_id"])
            self.assertNotEqual(query["trans_id"], self.panel._queries[0]["trans_id"])
            self.assertNotEqual(query["trans_id"], self.panel._queries[3]["trans_id"])
            self.assertTrue(query["starts_trans"])
            self.assertTrue(query["in_trans"])
            self.assertTrue(query["ends_trans"])

            query = self.panel._queries[5]
            self.assertEqual(query["alias"], "replica")
            self.assertIsNone(query["trans_id"])
            self.assertFalse("starts_trans" in query)
            self.assertFalse("in_trans" in query)
            self.assertFalse("end_trans" in query)
        else:
            # Ensure that nothing was recorded for other database engines.
            self.assertTrue(self.panel._queries)
            for query in self.panel._queries:
                self.assertFalse("trans_id" in query)
                self.assertFalse("starts_trans" in query)
                self.assertFalse("in_trans" in query)
                self.assertFalse("end_trans" in query)



from functools import cache

from django import template
from django.contrib.staticfiles import finders
from django.utils.html import format_html
from django.utils.safestring import mark_safe

register = template.Library()


@cache
def _svg_contents(path):
    absolute = finders.find(path)
    if absolute is None:
        return ""
    with open(absolute, encoding="utf-8") as handle:
        return handle.read().strip()


@register.simple_tag
def inline_svg(path, css_class="", label=""):
    """
    Inline a static SVG file instead of linking to it, avoiding a network
    request per icon. ``label`` names the icon for assistive tech; without one
    the icon is treated as decorative.
    """
    svg = _svg_contents(path)
    if not svg.startswith("<svg "):
        return ""
    if label:
        attrs = format_html('role="img" aria-label="{}"', label)
    else:
        attrs = mark_safe('aria-hidden="true" focusable="false"')
    if css_class:
        attrs = format_html('class="{}" {}', css_class, attrs)
    return mark_safe(svg.replace("<svg ", format_html("<svg {} ", attrs), 1))



"""Django settings for example project."""

import os
import sys

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production


SECRET_KEY = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"

DEBUG = True

INTERNAL_IPS = ["127.0.0.1", "::1"]

# Application definition

INSTALLED_APPS = [
    *(["daphne"] if os.getenv("ASYNC_SERVER", "") else []),
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "example.urls"

STATIC_URL = "/static/"

TEMPLATES = [
    {
        "NAME": "jinja2",
        "BACKEND": "django.template.backends.jinja2.Jinja2",
        "APP_DIRS": True,
        "DIRS": [os.path.join(BASE_DIR, "example", "templates", "jinja2")],
    },
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [os.path.join(BASE_DIR, "example", "templates")],
        "OPTIONS": {
            "debug": True,
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

USE_TZ = True

WSGI_APPLICATION = "example.wsgi.application"
ASGI_APPLICATION = "example.asgi.application"


# Cache and database

CACHES = {"default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"}}

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "example", "db.sqlite3"),
    }
}

# To use another database, set the DB_BACKEND environment variable.
if os.environ.get("DB_BACKEND", "").lower() == "postgresql":
    # See docs/contributing for instructions on configuring PostgreSQL.
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "NAME": "debug_toolbar",
            "USER": "debug_toolbar",
            "PASSWORD": "debug_toolbar",
        }
    }
if os.environ.get("DB_BACKEND", "").lower() == "mysql":
    # See docs/contributing for instructions on configuring MySQL/MariaDB.
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": "debug_toolbar",
            "USER": "debug_toolbar",
            "PASSWORD": "debug_toolbar",
        }
    }

STATICFILES_DIRS = [os.path.join(BASE_DIR, "example", "static")]

# Only enable the toolbar when we're in debug mode and we're
# not running tests. Django will change DEBUG to be False for
# tests, so we can't rely on DEBUG alone.
ENABLE_DEBUG_TOOLBAR = DEBUG and "test" not in sys.argv
if ENABLE_DEBUG_TOOLBAR:
    INSTALLED_APPS += [
        "debug_toolbar",
    ]
    MIDDLEWARE += [
        "debug_toolbar.middleware.DebugToolbarMiddleware",
    ]
    # Customize the config to support turbo and htmx boosting.
    DEBUG_TOOLBAR_CONFIG = {"ROOT_TAG_EXTRA_ATTRS": "data-turbo-permanent hx-preserve"}



import unittest

from debug_toolbar.sanitize import force_str


class ForceStrTestCase(unittest.TestCase):
    def test_success_convert(self):
        input = 0

        self.assertEqual(force_str(input), "0")

    def test_failed_convert(self):
        input = bytes.fromhex(
            "a3f2b8c14e972d5a8fb3c7291a64e0859c472bf63d18a0945e73b2c84f917ae2"
        )
        self.assertEqual(
            force_str(input), "Django Debug Toolbar was unable to parse value."
        )



import asyncio

import django
from asgiref.sync import sync_to_async
from django.contrib.auth.models import User
from django.core.cache import cache
from django.http import JsonResponse
from django.shortcuts import render

TASKS_AVAILABLE = django.VERSION >= (6, 0)

if TASKS_AVAILABLE:
    from django.tasks import task

    @task
    def example_task(x=1, y=2):
        return x + y


def tasks_view(request):
    if TASKS_AVAILABLE:
        example_task.enqueue(1, y=2)
        example_task.enqueue(3, y=4)
    return render(request, "tasks.html", {"tasks_available": TASKS_AVAILABLE})


def increment(request):
    try:
        value = int(request.session.get("value", 0)) + 1
    except ValueError:
        value = 1
    request.session["value"] = value
    return JsonResponse({"value": value})


def jinja2_view(request):
    return render(request, "index.jinja", {"foo": "bar"}, using="jinja2")


async def async_home(request):
    return await sync_to_async(render)(request, "index.html")


async def async_db(request):
    user_count = await User.objects.acount()

    return await sync_to_async(render)(
        request, "async_db.html", {"user_count": user_count}
    )


async def async_db_concurrent(request):
    # Do database queries concurrently
    (user_count, _) = await asyncio.gather(
        User.objects.acount(), User.objects.filter(username="test").acount()
    )

    return await sync_to_async(render)(
        request, "async_db.html", {"user_count": user_count}
    )


def cache_view(request):
    cache.set("foo", "bar")
    cache.get("foo")
    cache.get("baz")
    return render(request, "cache.html")



import sys
from unittest.mock import patch

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test import AsyncRequestFactory, RequestFactory, TestCase, override_settings

from debug_toolbar.middleware import DebugToolbarMiddleware

if sys.version_info >= (3, 12):
    from inspect import iscoroutinefunction
else:
    from asyncio import iscoroutinefunction


def show_toolbar_if_staff(request):
    # Hit the database, but always return True
    return User.objects.exists() or True


async def ashow_toolbar_if_staff(request):
    # Hit the database, but always return True
    has_users = await User.objects.afirst()
    return has_users or True


class MiddlewareSyncAsyncCompatibilityTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.async_factory = AsyncRequestFactory()

    @override_settings(DEBUG=True)
    def test_sync_mode(self):
        """
        test middleware switches to sync (__call__) based on get_response type
        """

        request = self.factory.get("/")
        middleware = DebugToolbarMiddleware(
            lambda x: HttpResponse("<html><body>Test app</body></html>")
        )

        self.assertFalse(iscoroutinefunction(middleware))

        response = middleware(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b"djdt", response.content)

    @override_settings(DEBUG=True)
    async def test_async_mode(self):
        """
        test middleware switches to async (__acall__) based on get_response type
        and returns a coroutine
        """

        async def get_response(request):
            return HttpResponse("<html><body>Test app</body></html>")

        middleware = DebugToolbarMiddleware(get_response)
        request = self.async_factory.get("/")

        self.assertTrue(iscoroutinefunction(middleware))

        response = await middleware(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b"djdt", response.content)

    @override_settings(DEBUG=True)
    @patch(
        "debug_toolbar.middleware.show_toolbar_func_or_path",
        return_value=ashow_toolbar_if_staff,
    )
    def test_async_show_toolbar_callback_sync_middleware(self, mocked_show):
        def get_response(request):
            return HttpResponse("<html><body>Hello world</body></html>")

        middleware = DebugToolbarMiddleware(get_response)

        request = self.factory.get("/")
        response = middleware(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b"djdt", response.content)

    @override_settings(DEBUG=True)
    @patch(
        "debug_toolbar.middleware.show_toolbar_func_or_path",
        return_value=show_toolbar_if_staff,
    )
    async def test_sync_show_toolbar_callback_async_middleware(self, mocked_show):
        async def get_response(request):
            return HttpResponse("<html><body>Hello world</body></html>")

        middleware = DebugToolbarMiddleware(get_response)

        request = self.async_factory.get("/")
        response = await middleware(request)
        self.assertEqual(response.status_code, 200)
        self.assertIn(b"djdt", response.content)



import sys
import unittest

from django.contrib.auth.models import User
from django.db import IntegrityError, transaction
from django.http import HttpResponse
from django.test.utils import override_settings

from debug_toolbar.panels.profiling import ProfilingPanel

from ..base import BaseTestCase, IntegrationTestCase
from ..views import listcomp_view, regular_view


@override_settings(
    DEBUG_TOOLBAR_PANELS=["debug_toolbar.panels.profiling.ProfilingPanel"]
)
class ProfilingPanelTestCase(BaseTestCase):
    panel_id = ProfilingPanel.panel_id

    def test_regular_view(self):
        self._get_response = lambda request: regular_view(request, "profiling")
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIn("func_list", self.panel.get_stats())
        self.assertIn("regular_view", self.panel.content)

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        self._get_response = lambda request: regular_view(request, "profiling")
        response = self.panel.process_request(self.request)
        # ensure the panel does not have content yet.
        self.assertNotIn("regular_view", self.panel.content)
        self.panel.generate_stats(self.request, response)
        self.reload_stats()
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn("regular_view", content)
        self.assertIn("render", content)
        self.assertValidHTML(content)
        # ensure traces aren't escaped
        self.assertIn('<span class="djdt-path">', content)

    @override_settings(DEBUG_TOOLBAR_CONFIG={"PROFILER_THRESHOLD_RATIO": 1})
    def test_cum_time_threshold(self):
        """
        Test that cumulative time threshold excludes calls
        """
        self._get_response = lambda request: regular_view(request, "profiling")
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel renders but doesn't include our function.
        content = self.panel.content
        self.assertIn("regular_view", content)
        self.assertNotIn("render", content)
        self.assertValidHTML(content)

    @unittest.skipUnless(
        sys.version_info < (3, 12, 0),
        "Python 3.12 no longer contains a frame for list comprehensions.",
    )
    def test_listcomp_escaped(self):
        self._get_response = lambda request: listcomp_view(request)
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        content = self.panel.content
        self.assertNotIn('<span class="djdt-func"><listcomp></span>', content)
        self.assertIn('<span class="djdt-func">&lt;listcomp&gt;</span>', content)

    def test_generate_stats_no_profiler(self):
        """
        Test generating stats with no profiler.
        """
        response = HttpResponse()
        self.assertIsNone(self.panel.generate_stats(self.request, response))

    def test_generate_stats_no_root_func(self):
        """
        Test generating stats using profiler without root function.
        """
        response = self.panel.process_request(self.request)
        self.panel.profiler.clear()
        self.panel.profiler.enable()
        self.panel.profiler.disable()
        self.panel.generate_stats(self.request, response)
        self.assertNotIn("func_list", self.panel.get_stats())


@override_settings(
    DEBUG=True, DEBUG_TOOLBAR_PANELS=["debug_toolbar.panels.profiling.ProfilingPanel"]
)
class ProfilingPanelIntegrationTestCase(IntegrationTestCase):
    def test_view_executed_once(self):
        self.assertEqual(User.objects.count(), 0)

        response = self.client.get("/new_user/")
        self.assertContains(response, "Profiling")
        self.assertEqual(User.objects.count(), 1)

        with self.assertRaises(IntegrityError), transaction.atomic():
            response = self.client.get("/new_user/")
        self.assertEqual(User.objects.count(), 1)



import copy
import html

from django.test import RequestFactory, override_settings
from django.urls import resolve, reverse

from debug_toolbar.panels.history import HistoryPanel
from debug_toolbar.panels.redirects import RedirectsPanel
from debug_toolbar.store import get_store
from debug_toolbar.toolbar import DebugToolbar

from ..base import BaseTestCase, IntegrationTestCase

rf = RequestFactory()


class HistoryPanelTestCase(BaseTestCase):
    panel_id = HistoryPanel.panel_id

    def test_disabled(self):
        config = {"DISABLE_PANELS": {"debug_toolbar.panels.history.HistoryPanel"}}
        self.assertTrue(self.panel.enabled)
        with self.settings(DEBUG_TOOLBAR_CONFIG=config):
            self.assertFalse(self.panel.enabled)

    def test_post(self):
        self.request = rf.post("/", data={"foo": "bar"})
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        data = self.panel.get_stats()["data"]
        self.assertEqual(data["foo"], "bar")

    def test_post_json(self):
        for data, expected_stats_data in (
            ({"foo": "bar"}, {"foo": "bar"}),
            ("", {}),  # Empty JSON
            ("'", {}),  # Invalid JSON
        ):
            with self.subTest(data=data):
                self.request = rf.post(
                    "/",
                    data=data,
                    content_type="application/json",
                    CONTENT_TYPE="application/json",  # Force django test client to add the content-type even if no data
                )
                response = self.panel.process_request(self.request)
                self.panel.generate_stats(self.request, response)
                data = self.panel.get_stats()["data"]
                self.assertDictEqual(data, expected_stats_data)

    def test_urls(self):
        self.assertEqual(
            reverse("djdt:history_sidebar"),
            "/__debug__/history_sidebar/",
        )
        self.assertEqual(
            resolve("/__debug__/history_sidebar/").url_name,
            "history_sidebar",
        )
        self.assertEqual(
            reverse("djdt:history_refresh"),
            "/__debug__/history_refresh/",
        )
        self.assertEqual(
            resolve("/__debug__/history_refresh/").url_name,
            "history_refresh",
        )


@override_settings(DEBUG=True)
class HistoryViewsTestCase(IntegrationTestCase):
    PANEL_KEYS = {
        "VersionsPanel",
        "TimerPanel",
        "SettingsPanel",
        "HeadersPanel",
        "RequestPanel",
        "SQLPanel",
        "StaticFilesPanel",
        "TemplatesPanel",
        "AlertsPanel",
        "CachePanel",
        "SignalsPanel",
        "TasksPanel",
        "CommunityPanel",
        "ProfilingPanel",
    }

    def test_history_panel_integration_content(self):
        """Verify the history panel's content renders properly.."""
        store = get_store()
        self.assertEqual(len(list(store.request_ids())), 0)

        data = {"foo": "bar"}
        self.client.get("/json_view/", data, content_type="application/json")

        # Check the history panel's stats to verify the toolbar rendered properly.
        request_ids = list(store.request_ids())
        self.assertEqual(len(request_ids), 1)
        toolbar = DebugToolbar.fetch(request_ids[0])
        content = toolbar.get_panel_by_id(HistoryPanel.panel_id).content
        self.assertIn("bar", content)
        self.assertIn('name="exclude_history" value="True"', content)

    def test_history_sidebar_invalid(self):
        response = self.client.get(reverse("djdt:history_sidebar"))
        self.assertEqual(response.status_code, 400)

    def test_history_headers(self):
        """Validate the headers injected from the history panel."""
        DebugToolbar.get_observe_request.cache_clear()
        response = self.client.get("/json_view/")
        request_id = next(iter(get_store().request_ids()))
        self.assertEqual(response.headers["djdt-request-id"], request_id)

    def test_history_headers_unobserved(self):
        """Validate the headers aren't injected from the history panel."""
        with self.settings(
            DEBUG_TOOLBAR_CONFIG={"OBSERVE_REQUEST_CALLBACK": lambda request: False}
        ):
            DebugToolbar.get_observe_request.cache_clear()
            response = self.client.get("/json_view/")
            self.assertNotIn("djdt-request-id", response.headers)
        # Clear it again to avoid conflicting with another test
        # Specifically, DebugToolbarLiveTestCase.test_ajax_refresh
        DebugToolbar.get_observe_request.cache_clear()

    def test_history_sidebar(self):
        """Validate the history sidebar view."""
        self.client.get("/json_view/")
        request_id = next(iter(get_store().request_ids()))
        data = {"request_id": request_id, "exclude_history": True}
        response = self.client.get(reverse("djdt:history_sidebar"), data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            set(response.json()),
            self.PANEL_KEYS,
        )

    def test_history_sidebar_includes_history(self):
        """Validate the history sidebar view."""
        self.client.get("/json_view/")
        panel_keys = copy.copy(self.PANEL_KEYS)
        panel_keys.add(HistoryPanel.panel_id)
        panel_keys.add(RedirectsPanel.panel_id)
        request_id = next(iter(get_store().request_ids()))
        data = {"request_id": request_id}
        response = self.client.get(reverse("djdt:history_sidebar"), data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            set(response.json()),
            panel_keys,
        )

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={"RENDER_PANELS": False, "RESULTS_CACHE_SIZE": 1}
    )
    def test_history_sidebar_expired_request_id(self):
        """Validate the history sidebar view."""
        self.client.get("/json_view/")
        request_id = next(iter(get_store().request_ids()))
        data = {"request_id": request_id, "exclude_history": True}
        response = self.client.get(reverse("djdt:history_sidebar"), data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            set(response.json()),
            self.PANEL_KEYS,
        )
        # Make enough requests to unset the original
        self.client.get("/json_view/")

        # Querying old request_id should return in empty response
        data = {"request_id": request_id, "exclude_history": True}
        response = self.client.get(reverse("djdt:history_sidebar"), data=data)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {})

        # Querying with latest request_id
        latest_request_id = next(iter(get_store().request_ids()))
        data = {"request_id": latest_request_id, "exclude_history": True}
        response = self.client.get(reverse("djdt:history_sidebar"), data=data)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            set(response.json()),
            self.PANEL_KEYS,
        )

    def test_history_refresh(self):
        """Verify refresh history response has request variables."""
        self.client.get("/json_view/", {"foo": "bar"}, content_type="application/json")
        self.client.get(
            "/json_view/", {"spam": "eggs"}, content_type="application/json"
        )

        response = self.client.get(
            reverse("djdt:history_refresh"), data={"request_id": "foo"}
        )
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(len(data["requests"]), 2)

        request_ids = list(get_store().request_ids())
        self.assertIn(html.escape(request_ids[0]), data["requests"][0]["content"])
        self.assertIn(html.escape(request_ids[1]), data["requests"][1]["content"])

        for val in ["foo", "bar"]:
            self.assertIn(val, data["requests"][0]["content"])

        for val in ["spam", "eggs"]:
            self.assertIn(val, data["requests"][1]["content"])



import sys

import django
from django.apps import apps
from django.utils.translation import gettext_lazy as _

from debug_toolbar.panels import Panel


class VersionsPanel(Panel):
    """
    Shows versions of Python, Django, and installed apps if possible.
    """

    is_async = True

    @property
    def nav_subtitle(self):
        return "Django %s" % self.get_stats()["django_version"]

    title = _("Versions")
    nav_icon = "debug_toolbar/img/versions.svg"

    template = "debug_toolbar/panels/versions.html"

    def generate_stats(self, request, response):
        versions = [
            ("Python", "", "%d.%d.%d" % sys.version_info[:3]),
            ("Django", "", self.get_app_version(django)),
        ]
        versions += list(self.gen_app_versions())
        self.record_stats(
            {
                "django_version": django.get_version(),
                "versions": sorted(versions, key=lambda v: v[0]),
                "paths": sys.path,
            }
        )

    def gen_app_versions(self):
        for app_config in apps.get_app_configs():
            name = app_config.verbose_name
            app = app_config.module
            version = self.get_app_version(app)
            if version:
                yield app.__name__, name, version

    def get_app_version(self, app):
        version = self.get_version_from_app(app)
        if isinstance(version, (list, tuple)):
            # We strip dots from the right because we do not want to show
            # trailing dots if there are empty elements in the list/tuple
            version = ".".join(str(o) for o in version).rstrip(".")
        return version

    def get_version_from_app(self, app):
        if hasattr(app, "get_version"):
            get_version = app.get_version
            if callable(get_version):
                try:
                    return get_version()
                except TypeError:
                    pass
            else:
                return get_version
        if hasattr(app, "VERSION"):
            return app.VERSION
        if hasattr(app, "__version__"):
            return app.__version__
        return



from django.test import override_settings

from debug_toolbar.panels import Panel

from ..base import IntegrationTestCase


class CustomPanel(Panel):
    def title(self):
        return "Title with special chars &\"'<>"


@override_settings(
    DEBUG=True, DEBUG_TOOLBAR_PANELS=["tests.panels.test_custom.CustomPanel"]
)
class CustomPanelTestCase(IntegrationTestCase):
    def test_escapes_panel_title(self):
        response = self.client.get("/regular/basic/")
        self.assertContains(
            response,
            """
            <li id="djdt-CustomPanel" class="djDebugPanelButton">
            <input type="checkbox" checked title="Disable for next and successive requests" data-cookie="djdtCustomPanel">
            <a class="CustomPanel" href="#" title="Title with special chars &amp;&quot;&#39;&lt;&gt;" aria-controls="CustomPanel" aria-expanded="false">
            Title with special chars &amp;&quot;&#39;&lt;&gt;
            </a>
            </li>
            """,
            html=True,
        )
        self.assertContains(
            response,
            """
            <div id="CustomPanel" class="djdt-panelContent djdt-hidden">
            <div class="djDebugPanelTitle">
            <h3>Title with special chars &amp;&quot;&#39;&lt;&gt;</h3>
            <button type="button" class="djDebugClose">×</button>
            </div>
            <div class="djDebugPanelContent">
            <div class="djdt-loader"></div>
            <div class="djdt-scroll" tabindex="0"></div>
            </div>
            </div>
            """,
            html=True,
        )



import contextlib
import uuid
from contextvars import ContextVar
from os.path import join, normpath

from django.contrib.staticfiles import finders, storage
from django.dispatch import Signal
from django.utils.translation import gettext_lazy as _, ngettext

from debug_toolbar import panels

# This will record and map the StaticFile instances with its associated
# request across threads and async concurrent requests state.
request_id_context_var = ContextVar("djdt_request_id_store")
record_static_file_signal = Signal()


class URLMixin:
    def url(self, path, *args, **kwargs):
        url = super().url(path, *args, **kwargs)
        with contextlib.suppress(LookupError):
            # For LookupError:
            # The ContextVar wasn't set yet. Since the toolbar wasn't properly
            # configured to handle this request, we don't need to capture
            # the static file.
            request_id = request_id_context_var.get()
            record_static_file_signal.send(
                sender=self,
                staticfile=(str(path), url, finders.find(str(path))),
                request_id=request_id,
            )
        return url


class StaticFilesPanel(panels.Panel):
    """
    A panel to display the found staticfiles.
    """

    is_async = True
    name = "Static files"
    template = "debug_toolbar/panels/staticfiles.html"

    @property
    def title(self):
        stats = self.get_stats()
        return _("Static files (%(num_found)s found, %(num_used)s used)") % {
            "num_found": stats.get("num_found"),
            "num_used": stats.get("num_used"),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_found = 0
        self.used_paths = set()
        self.request_id = str(uuid.uuid4())

    @classmethod
    def ready(cls):
        klass = storage.staticfiles_storage.__class__
        if URLMixin not in klass.mro():
            klass.__bases__ = (URLMixin, *klass.__bases__)

    def _store_static_files_signal_handler(self, sender, staticfile, **kwargs):
        # Only record the static file if the request_id matches the one
        # that was used to create the panel.
        # as sender of the signal and this handler will have multiple
        # concurrent connections and we want to avoid storing of same
        # staticfile from other connections as well.
        if request_id_context_var.get() == self.request_id:
            self.used_paths.add(staticfile)

    def enable_instrumentation(self):
        self.ctx_token = request_id_context_var.set(self.request_id)
        record_static_file_signal.connect(self._store_static_files_signal_handler)

    def disable_instrumentation(self):
        record_static_file_signal.disconnect(self._store_static_files_signal_handler)
        request_id_context_var.reset(self.ctx_token)

    nav_title = _("Static files")
    nav_icon = "debug_toolbar/img/staticfiles.svg"

    @property
    def nav_subtitle(self):
        num_used = self.get_stats().get("num_used")
        return ngettext(
            "%(num_used)s file used", "%(num_used)s files used", num_used
        ) % {"num_used": num_used}

    def generate_stats(self, request, response):
        self.record_stats(
            {
                "num_found": self.num_found,
                "num_used": len(self.used_paths),
                "staticfiles": sorted(self.used_paths),
                "staticfiles_apps": self.get_staticfiles_apps(),
                "staticfiles_dirs": self.get_staticfiles_dirs(),
                "staticfiles_finders": self.get_staticfiles_finders(),
            }
        )

    def get_staticfiles_finders(self):
        """
        Returns a sorted mapping between the finder path and the list
        of relative and file system paths which that finder was able
        to find.
        """
        finders_mapping = {}
        for finder in finders.get_finders():
            try:
                for path, finder_storage in finder.list([]):
                    if getattr(finder_storage, "prefix", None):
                        prefixed_path = join(finder_storage.prefix, path)
                    else:
                        prefixed_path = path
                    finder_cls = finder.__class__
                    finder_path = f"{finder_cls.__module__}.{finder_cls.__name__}"
                    real_path = finder_storage.path(path)
                    payload = (prefixed_path, real_path)
                    finders_mapping.setdefault(finder_path, []).append(payload)
                    self.num_found += 1
            except OSError:
                # This error should be captured and presented as a part of run_checks.
                pass
        return finders_mapping

    def get_staticfiles_dirs(self):
        """
        Returns a list of paths to inspect for additional static files
        """
        dirs = []
        for finder in finders.get_finders():
            if isinstance(finder, finders.FileSystemFinder):
                dirs.extend(finder.locations)
        return [(prefix, normpath(dir)) for prefix, dir in dirs]

    def get_staticfiles_apps(self):
        """
        Returns a list of app paths that have a static directory
        """
        apps = []
        for finder in finders.get_finders():
            if isinstance(finder, finders.AppDirectoriesFinder):
                for app in finder.apps:
                    if app not in apps:
                        apps.append(app)
        return apps



from django.http import HttpResponse, StreamingHttpResponse
from django.template import Context, Template

from ..base import BaseTestCase


class AlertsPanelTestCase(BaseTestCase):
    panel_id = "AlertsPanel"

    def test_alert_warning_display(self):
        """
        Test that the panel (does not) display[s] an alert when there are
        (no) problems.
        """
        self.panel.record_stats({"alerts": []})
        self.assertNotIn("alerts", self.panel.nav_subtitle)

        self.panel.record_stats({"alerts": ["Alert 1", "Alert 2"]})
        self.assertIn("2 alerts", self.panel.nav_subtitle)

    def test_file_form_without_enctype_multipart_form_data(self):
        """
        Test that the panel displays a form invalid message when there is
        a file input but encoding not set to multipart/form-data.
        """
        test_form = '<form id="test-form"><input type="file"></form>'
        result = self.panel.check_invalid_file_form_configuration(test_form)
        expected_error = (
            'Form with id "test-form" contains file input, '
            'but does not have the attribute enctype="multipart/form-data".'
        )
        self.assertEqual(result[0]["alert"], expected_error)
        self.assertEqual(len(result), 1)

    def test_file_form_no_id_without_enctype_multipart_form_data(self):
        """
        Test that the panel displays a form invalid message when there is
        a file input but encoding not set to multipart/form-data.

        This should use the message when the form has no id.
        """
        test_form = '<form><input type="file"></form>'
        result = self.panel.check_invalid_file_form_configuration(test_form)
        expected_error = (
            "Form contains file input, but does not have "
            'the attribute enctype="multipart/form-data".'
        )
        self.assertEqual(result[0]["alert"], expected_error)
        self.assertEqual(len(result), 1)

    def test_file_form_with_enctype_multipart_form_data(self):
        test_form = """<form id="test-form" enctype="multipart/form-data">
        <input type="file">
        </form>"""
        result = self.panel.check_invalid_file_form_configuration(test_form)

        self.assertEqual(len(result), 0)

    def test_file_form_with_enctype_multipart_form_data_in_button(self):
        test_form = """<form id="test-form">
        <input type="file">
        <input type="submit" formenctype="multipart/form-data">
        </form>"""
        result = self.panel.check_invalid_file_form_configuration(test_form)

        self.assertEqual(len(result), 0)

    def test_referenced_file_input_without_enctype_multipart_form_data(self):
        test_file_input = """<form id="test-form"></form>
        <input type="file" form = "test-form">"""
        result = self.panel.check_invalid_file_form_configuration(test_file_input)

        expected_error = (
            'Input element references form with id "test-form", '
            'but the form does not have the attribute enctype="multipart/form-data".'
        )
        self.assertEqual(result[0]["alert"], expected_error)
        self.assertEqual(len(result), 1)

    def test_referenced_file_input_with_enctype_multipart_form_data(self):
        test_file_input = """<form id="test-form" enctype="multipart/form-data">
        </form>
        <input type="file" form = "test-form">"""
        result = self.panel.check_invalid_file_form_configuration(test_file_input)

        self.assertEqual(len(result), 0)

    def test_integration_file_form_without_enctype_multipart_form_data(self):
        t = Template('<form id="test-form"><input type="file"></form>')
        c = Context({})
        rendered_template = t.render(c)
        response = HttpResponse(content=rendered_template)

        self.panel.generate_stats(self.request, response)

        self.assertIn("1 alert", self.panel.nav_subtitle)
        self.assertIn(
            "Form with id &quot;test-form&quot; contains file input, "
            "but does not have the attribute enctype=&quot;multipart/form-data&quot;.",
            self.panel.content,
        )

    def test_streaming_response(self):
        """Test to check for a streaming response."""

        def _render():
            yield "ok"

        response = StreamingHttpResponse(_render())

        self.panel.generate_stats(self.request, response)
        self.assertEqual(self.panel.get_stats(), {"alerts": []})



import unittest

from django.http import QueryDict
from django.test import override_settings

import debug_toolbar.utils
from debug_toolbar.utils import (
    get_name_from_obj,
    get_stack,
    get_stack_trace,
    render_stacktrace,
    sanitize_and_sort_request_vars,
    tidy_stacktrace,
)


class GetNameFromObjTestCase(unittest.TestCase):
    def test_func(self):
        def x():
            return 1

        res = get_name_from_obj(x)
        self.assertEqual(
            res, "tests.test_utils.GetNameFromObjTestCase.test_func.<locals>.x"
        )

    def test_lambda(self):
        res = get_name_from_obj(lambda: 1)
        self.assertEqual(
            res, "tests.test_utils.GetNameFromObjTestCase.test_lambda.<locals>.<lambda>"
        )

    def test_class(self):
        class A:
            pass

        res = get_name_from_obj(A)
        self.assertEqual(
            res, "tests.test_utils.GetNameFromObjTestCase.test_class.<locals>.A"
        )


class RenderStacktraceTestCase(unittest.TestCase):
    def test_importlib_path_issue_1612(self):
        trace = [
            ("/server/app.py", 1, "foo", ["code line 1", "code line 2"], {"foo": "bar"})
        ]
        result = render_stacktrace(trace)
        self.assertIn('<span class="djdt-path">/server/</span>', result)
        self.assertIn('<span class="djdt-file">app.py</span> in', result)

        trace = [
            (
                "<frozen importlib._bootstrap>",
                1,
                "foo",
                ["code line 1", "code line 2"],
                {"foo": "bar"},
            )
        ]
        result = render_stacktrace(trace)
        self.assertIn('<span class="djdt-path"></span>', result)
        self.assertIn(
            '<span class="djdt-file">&lt;frozen importlib._bootstrap&gt;</span> in',
            result,
        )


class StackTraceTestCase(unittest.TestCase):
    @override_settings(DEBUG_TOOLBAR_CONFIG={"HIDE_IN_STACKTRACES": []})
    def test_get_stack_trace_skip(self):
        stack_trace = get_stack_trace(skip=-1)
        self.assertTrue(len(stack_trace) > 2)
        self.assertEqual(stack_trace[-1][0], debug_toolbar.utils.__file__)
        self.assertEqual(stack_trace[-1][2], "get_stack_trace")
        self.assertEqual(stack_trace[-2][0], __file__)
        self.assertEqual(stack_trace[-2][2], "test_get_stack_trace_skip")

        stack_trace = get_stack_trace()
        self.assertTrue(len(stack_trace) > 1)
        self.assertEqual(stack_trace[-1][0], __file__)
        self.assertEqual(stack_trace[-1][2], "test_get_stack_trace_skip")

    def test_deprecated_functions(self):
        with self.assertWarns(DeprecationWarning):
            stack = get_stack()
        self.assertEqual(stack[0][1], __file__)
        with self.assertWarns(DeprecationWarning):
            stack_trace = tidy_stacktrace(reversed(stack))
        self.assertEqual(stack_trace[-1][0], __file__)

    @override_settings(DEBUG_TOOLBAR_CONFIG={"ENABLE_STACKTRACES_LOCALS": True})
    def test_locals(self):
        # This wrapper class is necessary to mask the repr() of the list
        # returned by get_stack_trace(); otherwise the 'test_locals_value_1'
        # string will also be present in rendered_stack_2.
        class HideRepr:
            def __init__(self, value):
                self.value = value

        x = "test_locals_value_1"
        stack_1_wrapper = HideRepr(get_stack_trace())

        x = x.replace("1", "2")
        stack_2_wrapper = HideRepr(get_stack_trace())

        rendered_stack_1 = render_stacktrace(stack_1_wrapper.value)
        self.assertIn("test_locals_value_1", rendered_stack_1)
        self.assertNotIn("test_locals_value_2", rendered_stack_1)

        rendered_stack_2 = render_stacktrace(stack_2_wrapper.value)
        self.assertNotIn("test_locals_value_1", rendered_stack_2)
        self.assertIn("test_locals_value_2", rendered_stack_2)


class SanitizeAndSortRequestVarsTestCase(unittest.TestCase):
    """Tests for the sanitize_and_sort_request_vars function."""

    def test_dict_sanitization(self):
        """Test sanitization of a regular dictionary."""
        test_dict = {
            "username": "testuser",
            "password": "secret123",
            "api_key": "abc123",
        }
        result = sanitize_and_sort_request_vars(test_dict)

        # Convert to dict for easier testing
        result_dict = dict(result["list"])

        self.assertEqual(result_dict["username"], "testuser")
        self.assertEqual(result_dict["password"], "********************")
        self.assertEqual(result_dict["api_key"], "********************")

    def test_querydict_sanitization(self):
        """Test sanitization of a QueryDict."""
        query_dict = QueryDict("username=testuser&password=secret123&api_key=abc123")
        result = sanitize_and_sort_request_vars(query_dict)

        # Convert to dict for easier testing
        result_dict = dict(result["list"])

        self.assertEqual(result_dict["username"], "testuser")
        self.assertEqual(result_dict["password"], "********************")
        self.assertEqual(result_dict["api_key"], "********************")

    def test_non_sortable_dict_keys(self):
        """Test dictionary with keys that can't be sorted."""
        test_dict = {
            1: "one",
            "2": "two",
            None: "none",
        }
        result = sanitize_and_sort_request_vars(test_dict)
        self.assertEqual(len(result["list"]), 3)
        result_dict = dict(result["list"])
        self.assertEqual(result_dict[1], "one")
        self.assertEqual(result_dict["2"], "two")
        self.assertEqual(result_dict[None], "none")

    def test_querydict_multiple_values(self):
        """Test QueryDict with multiple values for the same key."""
        query_dict = QueryDict("name=bar1&name=bar2&title=value")
        result = sanitize_and_sort_request_vars(query_dict)
        result_dict = dict(result["list"])
        self.assertEqual(result_dict["name"], ["bar1", "bar2"])
        self.assertEqual(result_dict["title"], "value")

    def test_non_dict_input(self):
        """Test handling of non-dict input."""
        test_input = ["not", "a", "dict"]
        result = sanitize_and_sort_request_vars(test_input)
        self.assertEqual(result["raw"], test_input)



from pathlib import Path

from django.conf import settings
from django.contrib.staticfiles import finders, storage
from django.shortcuts import render
from django.test import AsyncRequestFactory, RequestFactory

from debug_toolbar.panels.staticfiles import StaticFilesPanel, URLMixin

from ..base import BaseTestCase


class StaticFilesPanelTestCase(BaseTestCase):
    panel_id = StaticFilesPanel.panel_id

    def test_default_case(self):
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        content = self.panel.content
        self.assertIn(
            "django.contrib.staticfiles.finders.AppDirectoriesFinder", content
        )
        self.assertIn(
            "django.contrib.staticfiles.finders.FileSystemFinder (2 files)", content
        )
        self.assertEqual(self.panel.get_stats()["num_used"], 0)
        self.assertNotEqual(self.panel.num_found, 0)
        expected_apps = ["django.contrib.admin", "debug_toolbar"]
        if settings.USE_GIS:
            expected_apps = ["django.contrib.gis"] + expected_apps
        self.assertEqual(self.panel.get_staticfiles_apps(), expected_apps)
        self.assertEqual(
            self.panel.get_staticfiles_dirs(), finders.FileSystemFinder().locations
        )

    async def test_store_staticfiles_with_async_context(self):
        async def get_response(request):
            # template contains one static file
            return render(request, "staticfiles/async_static.html")

        self._get_response = get_response
        async_request = AsyncRequestFactory().get("/")
        response = await self.panel.process_request(async_request)
        self.panel.generate_stats(self.request, response)
        self.assertEqual(self.panel.get_stats()["num_used"], 1)

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        response = self.panel.process_request(self.request)
        # ensure the panel does not have content yet.
        self.assertNotIn(
            "django.contrib.staticfiles.finders.AppDirectoriesFinder",
            self.panel.content,
        )
        self.panel.generate_stats(self.request, response)
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn(
            "django.contrib.staticfiles.finders.AppDirectoriesFinder", content
        )
        self.assertValidHTML(content)

    def test_path(self):
        def get_response(request):
            return render(
                request,
                "staticfiles/path.html",
                {
                    "paths": [
                        Path("additional_static/base.css"),
                        "additional_static/base.css",
                        "additional_static/base2.css",
                    ]
                },
            )

        self._get_response = get_response
        request = RequestFactory().get("/")
        response = self.panel.process_request(request)
        self.panel.generate_stats(self.request, response)
        self.assertEqual(self.panel.get_stats()["num_used"], 2)
        self.assertIn(
            'href="/static/additional_static/base.css"', self.panel.content, 1
        )
        self.assertIn(
            'href="/static/additional_static/base2.css"', self.panel.content, 1
        )

    def test_storage_state_preservation(self):
        """Ensure the URLMixin doesn't affect storage state"""
        original_storage = storage.staticfiles_storage
        original_attrs = dict(original_storage.__dict__)

        # Trigger mixin injection
        self.panel.ready()

        # Verify all original attributes are preserved
        self.assertEqual(original_attrs, dict(original_storage.__dict__))

    def test_context_variable_lifecycle(self):
        """Test the request_id context variable lifecycle"""
        from debug_toolbar.panels.staticfiles import request_id_context_var

        # Should not raise when context not set
        url = storage.staticfiles_storage.url("test.css")
        self.assertTrue(url.startswith("/static/"))

        # Should track when context is set
        token = request_id_context_var.set("test-request-id")
        try:
            url = storage.staticfiles_storage.url("test.css")
            self.assertTrue(url.startswith("/static/"))
            # Verify file was tracked
            self.assertIn("test.css", [f[0] for f in self.panel.used_paths])
        finally:
            request_id_context_var.reset(token)

    def test_multiple_initialization(self):
        """Ensure multiple panel initializations don't stack URLMixin"""
        storage_class = storage.staticfiles_storage.__class__

        # Initialize panel multiple times
        for _ in range(3):
            self.panel.ready()

        # Verify URLMixin appears exactly once in bases
        mixin_count = sum(1 for base in storage_class.__bases__ if base == URLMixin)
        self.assertEqual(mixin_count, 1)



import io
import sys

from django.contrib.auth.models import User
from django.core import management
from django.db import connection
from django.test import TestCase
from django.test.utils import override_settings

if connection.vendor == "postgresql":
    from django.db.backends.postgresql import base as base_module
else:
    from django.db.backends import utils as base_module


@override_settings(DEBUG=True)
class DebugSQLShellTestCase(TestCase):
    def setUp(self):
        self.original_wrapper = base_module.CursorDebugWrapper
        # Since debugsqlshell monkey-patches django.db.backends.utils, we can
        # test it simply by loading it, without executing it. But we have to
        # undo the monkey-patch on exit.
        command_name = "debugsqlshell"
        app_name = management.get_commands()[command_name]
        management.load_command_class(app_name, command_name)

    def tearDown(self):
        base_module.CursorDebugWrapper = self.original_wrapper

    def test_command(self):
        original_stdout, sys.stdout = sys.stdout, io.StringIO()
        try:
            User.objects.count()
            self.assertIn("SELECT COUNT", sys.stdout.getvalue())
        finally:
            sys.stdout = original_stdout



from django.core import cache

from debug_toolbar.panels.cache import CachePanel

from ..base import BaseTestCase


class CachePanelTestCase(BaseTestCase):
    panel_id = CachePanel.panel_id

    def test_recording(self):
        self.assertEqual(len(self.panel.calls), 0)
        cache.cache.set("foo", "bar")
        cache.cache.get("foo")
        cache.cache.delete("foo")
        self.assertFalse(cache.cache.touch("foo"))
        cache.cache.set("foo", "bar")
        self.assertTrue(cache.cache.touch("foo"))
        # Verify that the cache has a valid clear method.
        cache.cache.clear()
        self.assertEqual(len(self.panel.calls), 7)

    def test_recording_caches(self):
        self.assertEqual(len(self.panel.calls), 0)
        default_cache = cache.caches[cache.DEFAULT_CACHE_ALIAS]
        second_cache = cache.caches["second"]
        default_cache.set("foo", "bar")
        second_cache.get("foo")
        self.assertEqual(len(self.panel.calls), 2)

    def test_hits_and_misses(self):
        cache.cache.clear()
        cache.cache.get("foo")
        self.assertEqual(self.panel.hits, 0)
        self.assertEqual(self.panel.misses, 1)
        cache.cache.set("foo", 1)
        cache.cache.get("foo")
        self.assertEqual(self.panel.hits, 1)
        self.assertEqual(self.panel.misses, 1)
        cache.cache.get_many(["foo", "bar"])
        self.assertEqual(self.panel.hits, 2)
        self.assertEqual(self.panel.misses, 2)
        cache.cache.set("bar", 2)
        cache.cache.get_many(keys=["foo", "bar"])
        self.assertEqual(self.panel.hits, 4)
        self.assertEqual(self.panel.misses, 2)

    def test_get_default_counts_as_miss(self):
        """
        Using get that fetches a cached value that's the same as the default
        results in a miss despite it being a hit.
        """
        cache.cache.clear()

        # cached None with default None is ambiguous
        cache.cache.set("foo", None)
        self.assertIsNone(cache.cache.get("foo"))
        self.assertIsNone(cache.cache.get("foo"))
        self.assertEqual(self.panel.misses, 2)
        self.assertEqual(self.panel.hits, 0)

        # cached value equal to default is ambiguous
        cache.cache.set("bar", "baz")
        self.assertEqual(cache.cache.get("bar", "baz"), "baz")
        self.assertEqual(cache.cache.get("bar", "baz"), "baz")
        self.assertEqual(self.panel.misses, 4)
        self.assertEqual(self.panel.hits, 0)

    def test_get_with_args(self):
        """
        Test get to calculate the default in a variety of ways.
        """
        cache.cache.clear()
        self.assertIsNone(cache.cache.get("foo"))
        self.assertEqual(self.panel.misses, 1)
        self.assertEqual(cache.cache.get("foo", "bar"), "bar")
        self.assertEqual(self.panel.misses, 2)
        self.assertEqual(cache.cache.get("foo", default="bar"), "bar")
        self.assertEqual(self.panel.misses, 3)
        self.assertEqual(cache.cache.get(key="foo", default="bar"), "bar")
        self.assertEqual(self.panel.misses, 4)
        self.assertEqual(self.panel.hits, 0)

    def test_get_or_set_value(self):
        cache.cache.get_or_set("baz", "val")
        self.assertEqual(cache.cache.get("baz"), "val")
        calls = [
            (call["name"], call["args"], call["kwargs"]) for call in self.panel.calls
        ]
        self.assertEqual(
            calls,
            [
                ("get_or_set", ("baz", "val"), {}),
                ("get", ("baz",), {}),
            ],
        )
        self.assertEqual(
            self.panel.counts,
            {
                "add": 0,
                "get": 1,
                "set": 0,
                "get_or_set": 1,
                "touch": 0,
                "delete": 0,
                "clear": 0,
                "get_many": 0,
                "set_many": 0,
                "delete_many": 0,
                "has_key": 0,
                "incr": 0,
                "decr": 0,
                "incr_version": 0,
                "decr_version": 0,
            },
        )

    def test_get_or_set_does_not_override_existing_value(self):
        cache.cache.set("foo", "bar")
        cached_value = cache.cache.get_or_set("foo", "other")
        self.assertEqual(cached_value, "bar")
        calls = [
            (call["name"], call["args"], call["kwargs"]) for call in self.panel.calls
        ]
        self.assertEqual(
            calls,
            [
                ("set", ("foo", "bar"), {}),
                ("get_or_set", ("foo", "other"), {}),
            ],
        )
        self.assertEqual(
            self.panel.counts,
            {
                "add": 0,
                "get": 0,
                "set": 1,
                "get_or_set": 1,
                "touch": 0,
                "delete": 0,
                "clear": 0,
                "get_many": 0,
                "set_many": 0,
                "delete_many": 0,
                "has_key": 0,
                "incr": 0,
                "decr": 0,
                "incr_version": 0,
                "decr_version": 0,
            },
        )

    def test_get_or_set_none_counts_as_miss(self):
        """
        Using get_or_set to set a None value is counted as a miss even
        though it was a hit on the second try.
        """
        cache.cache.clear()
        self.assertIsNone(cache.cache.get_or_set("foo", None))
        self.assertIsNone(cache.cache.get_or_set("foo", None))
        self.assertEqual(self.panel.misses, 2)
        self.assertEqual(self.panel.hits, 0)

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        cache.cache.get("café")
        response = self.panel.process_request(self.request)
        # ensure the panel does not have content yet.
        self.assertNotIn("café", self.panel.content)
        self.panel.generate_stats(self.request, response)
        self.reload_stats()
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn("café", content)
        self.assertValidHTML(content)
        # ensure traces aren't escaped
        self.assertIn('<span class="djdt-path">', content)

    def test_generate_server_timing(self):
        self.assertEqual(len(self.panel.calls), 0)
        cache.cache.set("foo", "bar")
        cache.cache.get("foo")
        cache.cache.delete("foo")

        self.assertEqual(len(self.panel.calls), 3)

        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.panel.generate_server_timing(self.request, response)

        stats = self.panel.get_stats()

        expected_data = {
            "total_time": {
                "title": "Cache {} Calls".format(stats["total_calls"]),
                "value": stats["total_time"],
            }
        }

        self.assertEqual(self.panel.get_server_timing_stats(), expected_data)

    def test_backend_alias_is_recorded(self):
        cache.cache.get("foo")
        self.assertEqual(self.panel.calls[0]["backend"], "default (LocMemCache)")

    def test_patching_only_applied_once(self):
        """
        Confirm that reapplying and disabling the cache patching only wraps
        the cache methods once
        """
        for _ in range(2):
            self.panel.enable_instrumentation()
            self.panel.disable_instrumentation()
        nested_wraps = 0
        wrapped = cache.cache.get.__wrapped__
        while wrapped is not None:
            try:
                wrapped = wrapped.__wrapped__
                nested_wraps += 1
            except AttributeError:
                break
        self.assertEqual(nested_wraps, 0)



from unittest.mock import patch

from django.apps import apps
from django.test import SimpleTestCase, override_settings

from debug_toolbar.apps import _manage_migrations_visibility


class AppsTestCase(SimpleTestCase):
    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore"
        }
    )
    @patch("debug_toolbar.apps.settings.MIGRATION_MODULES")
    def test_migrations_are_visible(self, mocked_migration_modules):
        _manage_migrations_visibility("debug_toolbar")
        self.assertFalse(mocked_migration_modules.setdefault.called)

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={"TOOLBAR_STORE_CLASS": "debug_toolbar.store.MemoryStore"}
    )
    @patch("debug_toolbar.apps.settings.MIGRATION_MODULES")
    def test_migrations_are_hidden(self, mocked_migration_modules):
        _manage_migrations_visibility("debug_toolbar")
        mocked_migration_modules.setdefault.assert_called_once_with(
            "debug_toolbar", None
        )

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={
            "TOOLBAR_STORE_CLASS": "debug_toolbar.store.DatabaseStore"
        }
    )
    def test_models_are_visible(self):
        app_config = apps.get_app_config("debug_toolbar")
        app_config.import_models()
        apps.get_model("debug_toolbar", "HistoryEntry")

    @override_settings(
        DEBUG_TOOLBAR_CONFIG={"TOOLBAR_STORE_CLASS": "debug_toolbar.store.MemoryStore"}
    )
    def test_models_are_hidden(self):
        app_config = apps.get_app_config("debug_toolbar")
        app_config.import_models()
        with self.assertRaises(LookupError):
            apps.get_model("debug_toolbar", "HistoryEntry")



from unittest.mock import patch

from django.test import TestCase

from debug_toolbar.settings import _is_running_tests


class SettingsTestCase(TestCase):
    @patch("debug_toolbar.settings.sys")
    @patch("debug_toolbar.settings.os")
    def test_is_running_tests(self, mock_os, mock_sys):
        mock_sys.argv = "test"
        mock_os.environ = {}
        self.assertTrue(_is_running_tests())

        mock_sys.argv = ""
        mock_os.environ = {}
        self.assertFalse(_is_running_tests())

        mock_sys.argv = ""
        mock_os.environ = {"PYTEST_VERSION": "1"}
        self.assertTrue(_is_running_tests())



from django.urls import path

from example.async_ import views
from example.urls import urlpatterns as sync_urlpatterns

urlpatterns = [
    path("async/db/", views.async_db_view, name="async_db_view"),
    *sync_urlpatterns,
]



from django.core import signing
from django.http import HttpResponseBadRequest, JsonResponse
from django.template import Origin, TemplateDoesNotExist
from django.template.engine import Engine
from django.template.loader import render_to_string
from django.utils.html import format_html, mark_safe

from debug_toolbar._compat import login_not_required
from debug_toolbar.decorators import render_with_toolbar_language, require_show_toolbar


@login_not_required
@require_show_toolbar
@render_with_toolbar_language
def template_source(request):
    """
    Return the source of a template, syntax-highlighted by Pygments if
    it's available.
    """
    template_origin_name = request.GET.get("template_origin")
    if template_origin_name is None:
        return HttpResponseBadRequest('"template_origin" key is required')
    try:
        template_origin_name = signing.loads(template_origin_name)
    except signing.BadSignature:
        return HttpResponseBadRequest('"template_origin" is invalid')
    template_name = request.GET.get("template", template_origin_name)

    final_loaders = []
    loaders = list(Engine.get_default().template_loaders)

    while loaders:
        loader = loaders.pop(0)

        if loader is not None:
            # Recursively unwrap loaders until we get to loaders which do not
            # themselves wrap other loaders. This adds support for
            # django.template.loaders.cached.Loader and the
            # django-template-partials loader (possibly among others)
            if hasattr(loader, "loaders"):
                loaders.extend(loader.loaders)
            else:
                final_loaders.append(loader)

    for loader in final_loaders:
        origin = Origin(template_origin_name)
        try:
            source = loader.get_contents(origin)
            break
        except TemplateDoesNotExist:
            pass
    else:
        source = f"Template Does Not Exist: {template_origin_name}"

    try:
        from pygments import highlight
        from pygments.formatters import HtmlFormatter
        from pygments.lexers import HtmlDjangoLexer
    except ModuleNotFoundError:
        source = format_html("<code>{}</code>", source)
    else:
        source = highlight(source, HtmlDjangoLexer(), HtmlFormatter(wrapcode=True))
        source = mark_safe(source)

    content = render_to_string(
        "debug_toolbar/panels/template_source.html",
        {"source": source, "template_name": template_name},
    )
    return JsonResponse({"content": content})



from __future__ import annotations

import inspect
import linecache
import os.path
import sys
import warnings
from collections.abc import Sequence
from pprint import PrettyPrinter, pformat
from typing import Any

from asgiref.local import Local
from django.http import QueryDict
from django.template import Node
from django.utils.html import format_html
from django.utils.safestring import SafeString, mark_safe
from django.views.debug import get_default_exception_reporter_filter

from debug_toolbar import _compat as compat, _stubs as stubs, settings as dt_settings

_local_data = Local()
safe_filter = get_default_exception_reporter_filter()


def _is_excluded_frame(frame: Any, excluded_modules: Sequence[str] | None) -> bool:
    if not excluded_modules:
        return False
    frame_module = frame.f_globals.get("__name__")
    if not isinstance(frame_module, str):
        return False
    return any(
        frame_module == excluded_module
        or frame_module.startswith(excluded_module + ".")
        for excluded_module in excluded_modules
    )


def _stack_trace_deprecation_warning() -> None:
    warnings.warn(
        "get_stack() and tidy_stacktrace() are deprecated in favor of"
        " get_stack_trace()",
        DeprecationWarning,
        stacklevel=2,
    )


def tidy_stacktrace(stack: list[stubs.InspectStack]) -> stubs.TidyStackTrace:
    """
    Clean up stacktrace and remove all entries that are excluded by the
    HIDE_IN_STACKTRACES setting.

    ``stack`` should be a list of frame tuples from ``inspect.stack()`` or
    ``debug_toolbar.utils.get_stack()``.
    """
    _stack_trace_deprecation_warning()

    trace = []
    excluded_modules = dt_settings.get_config()["HIDE_IN_STACKTRACES"]
    for frame, path, line_no, func_name, text in (f[:5] for f in stack):
        if _is_excluded_frame(frame, excluded_modules):
            continue
        text = "".join(text).strip() if text else ""
        frame_locals = (
            pformat(frame.f_locals)
            if dt_settings.get_config()["ENABLE_STACKTRACES_LOCALS"]
            else None
        )
        trace.append((path, line_no, func_name, text, frame_locals))
    return trace


def render_stacktrace(trace: stubs.TidyStackTrace) -> SafeString:
    show_locals = dt_settings.get_config()["ENABLE_STACKTRACES_LOCALS"]
    html = ""
    for abspath, lineno, func, code, locals_ in trace:
        if os.path.sep in abspath:
            directory, filename = abspath.rsplit(os.path.sep, 1)
            # We want the separator to appear in the UI so add it back.
            directory += os.path.sep
        else:
            # abspath could be something like "<frozen importlib._bootstrap>"
            directory = ""
            filename = abspath
        html += format_html(
            (
                '<span class="djdt-path">{}</span>'
                + '<span class="djdt-file">{}</span> in'
                + ' <span class="djdt-func">{}</span>'
                + '(<span class="djdt-lineno">{}</span>)\n'
                + '  <span class="djdt-code">{}</span>\n'
            ),
            directory,
            filename,
            func,
            lineno,
            code,
        )
        if show_locals:
            html += format_html(
                '  <pre class="djdt-locals">{}</pre>\n',
                locals_,
            )
        html += "\n"
    return mark_safe(html)


def get_template_info() -> dict[str, Any] | None:
    template_info = None
    cur_frame = sys._getframe().f_back
    try:
        while cur_frame is not None:
            in_utils_module = cur_frame.f_code.co_filename.endswith(
                "/debug_toolbar/utils.py"
            )
            is_get_template_context = (
                cur_frame.f_code.co_name == get_template_context.__name__
            )
            if in_utils_module and is_get_template_context:
                # If the method in the stack trace is this one
                # then break from the loop as it's being check recursively.
                break
            elif cur_frame.f_code.co_name == "render":
                node = cur_frame.f_locals["self"]
                context = cur_frame.f_locals["context"]
                if isinstance(node, Node):
                    template_info = get_template_context(node, context)
                    break
            cur_frame = cur_frame.f_back
    except Exception:  # noqa: S110, BLE001
        pass
    del cur_frame
    return template_info


def get_template_context(
    node: Node, context: stubs.RequestContext, context_lines: int = 3
) -> dict[str, Any]:
    line, source_lines, name = get_template_source_from_exception_info(node, context)
    debug_context = []
    start = max(1, line - context_lines)
    end = line + 1 + context_lines

    for line_num, content in source_lines:
        if start <= line_num <= end:
            debug_context.append(
                {"num": line_num, "content": content, "highlight": (line_num == line)}
            )

    return {"name": name, "context": debug_context}


def get_template_source_from_exception_info(
    node: Node, context: stubs.RequestContext
) -> tuple[int, list[tuple[int, str]], str]:
    if context.template.origin == node.origin:
        exception_info = context.template.get_exception_info(
            Exception("DDT"), node.token
        )
    else:
        exception_info = context.render_context.template.get_exception_info(
            Exception("DDT"), node.token
        )
    line = exception_info["line"]
    source_lines = exception_info["source_lines"]
    name = exception_info["name"]
    return line, source_lines, name


def get_name_from_obj(obj: Any) -> str:
    """Get the best name as `str` from a view or a object."""
    # This is essentially a rewrite of the `django.contrib.admindocs.utils.get_view_name`
    # https://github.com/django/django/blob/9a22d1769b042a88741f0ff3087f10d94f325d86/django/contrib/admindocs/utils.py#L26-L32
    if hasattr(obj, "view_class"):
        klass = obj.view_class
        return f"{klass.__module__}.{klass.__qualname__}"
    mod_name = obj.__module__
    view_name = getattr(obj, "__qualname__", obj.__class__.__name__)
    return mod_name + "." + view_name


def getframeinfo(frame: Any, context: int = 1) -> inspect.Traceback:
    """
    Get information about a frame or traceback object.

    A tuple of five things is returned: the filename, the line number of
    the current line, the function name, a list of lines of context from
    the source code, and the index of the current line within that list.
    The optional second argument specifies the number of lines of context
    to return, which are centered around the current line.

    This originally comes from ``inspect`` but is modified to handle issues
    with ``findsource()``.
    """
    if inspect.istraceback(frame):
        lineno = frame.tb_lineno
        frame = frame.tb_frame
    else:
        lineno = frame.f_lineno
    if not inspect.isframe(frame):
        raise TypeError("arg is not a frame or traceback object")

    filename = inspect.getsourcefile(frame) or inspect.getfile(frame)
    if context > 0:
        start = lineno - 1 - context // 2
        try:
            lines, _lnum = inspect.findsource(frame)
        except Exception:  # noqa: BLE001
            # findsource raises platform-dependant exceptions
            lines = index = None
        else:
            start = max(start, 1)
            start = max(0, min(start, len(lines) - context))
            lines = lines[start : (start + context)]
            index = lineno - 1 - start
    else:
        lines = index = None

    return inspect.Traceback(filename, lineno, frame.f_code.co_name, lines, index)


def sanitize_and_sort_request_vars(
    variable: dict[str, Any] | QueryDict,
) -> dict[str, list[tuple[str, Any]] | Any]:
    """
    Get a data structure for showing a sorted list of variables from the
    request data with sensitive values redacted.
    """
    if not isinstance(variable, (dict, QueryDict)):
        return {"raw": variable}

    # Get sorted keys if possible, otherwise just list them
    keys = _get_sorted_keys(variable)

    # Process the variable based on its type
    if isinstance(variable, QueryDict):
        result = _process_query_dict(variable, keys)
    else:
        result = _process_dict(variable, keys)

    return {"list": result}


def _get_sorted_keys(variable):
    """Helper function to get sorted keys if possible."""
    try:
        return sorted(variable)
    except TypeError:
        return list(variable)


def _process_query_dict(query_dict, keys):
    """Process a QueryDict into a list of (key, sanitized_value) tuples."""
    result = []
    for k in keys:
        values = query_dict.getlist(k)
        # Return single value if there's only one, otherwise keep as list
        value = values[0] if len(values) == 1 else values
        result.append((k, safe_filter.cleanse_setting(k, value)))
    return result


def _process_dict(dictionary, keys):
    """Process a dictionary into a list of (key, sanitized_value) tuples."""
    return [(k, safe_filter.cleanse_setting(k, dictionary.get(k))) for k in keys]


def get_stack(context=1) -> list[stubs.InspectStack]:
    """
    Get a list of records for a frame and all higher (calling) frames.

    Each record contains a frame object, filename, line number, function
    name, a list of lines of context, and index within the context.

    Modified version of ``inspect.stack()`` which calls our own ``getframeinfo()``
    """
    _stack_trace_deprecation_warning()

    frame = sys._getframe(1)
    framelist = []
    while frame:
        framelist.append((frame,) + getframeinfo(frame, context))
        frame = frame.f_back
    return framelist


def _stack_frames(*, skip=0):
    skip += 1  # Skip the frame for this generator.
    frame = inspect.currentframe()
    while frame is not None:
        if skip > 0:
            skip -= 1
        else:
            yield frame
        frame = frame.f_back


class _StackTraceRecorder:
    pretty_printer = PrettyPrinter()

    def __init__(self):
        self.filename_cache = {}

    def get_source_file(self, frame):
        frame_filename = frame.f_code.co_filename

        value = self.filename_cache.get(frame_filename)
        if value is None:
            filename = inspect.getsourcefile(frame)
            if filename is None:
                is_source = False
                filename = frame_filename
            else:
                is_source = True
                # Ensure linecache validity the first time this recorder
                # encounters the filename in this frame.
                linecache.checkcache(filename)
            value = (filename, is_source)
            self.filename_cache[frame_filename] = value

        return value

    def get_stack_trace(
        self,
        *,
        excluded_modules: Sequence[str] | None = None,
        include_locals: bool = False,
        skip: int = 0,
    ):
        trace = []
        skip += 1  # Skip the frame for this method.
        for frame in _stack_frames(skip=skip):
            if _is_excluded_frame(frame, excluded_modules):
                continue

            filename, is_source = self.get_source_file(frame)

            line_no = frame.f_lineno
            func_name = frame.f_code.co_name

            if is_source:
                module = inspect.getmodule(frame, filename)
                module_globals = module.__dict__ if module is not None else None
                source_line = linecache.getline(
                    filename, line_no, module_globals
                ).strip()
            else:
                source_line = ""

            if include_locals:
                frame_locals = self.pretty_printer.pformat(frame.f_locals)
            else:
                frame_locals = None

            trace.append((filename, line_no, func_name, source_line, frame_locals))
        trace.reverse()
        return trace


def get_stack_trace(*, skip=0):
    """
    Return a processed stack trace for the current call stack.

    If the ``ENABLE_STACKTRACES`` setting is False, return an empty :class:`list`.
    Otherwise return a :class:`list` of processed stack frame tuples (file name, line
    number, function name, source line, frame locals) for the current call stack.  The
    first entry in the list will be for the bottom of the stack and the last entry will
    be for the top of the stack.

    ``skip`` is an :class:`int` indicating the number of stack frames above the frame
    for this function to omit from the stack trace.  The default value of ``0`` means
    that the entry for the caller of this function will be the last entry in the
    returned stack trace.
    """
    config = dt_settings.get_config()
    if not config["ENABLE_STACKTRACES"]:
        return []
    skip += 1  # Skip the frame for this function.
    stack_trace_recorder = getattr(_local_data, "stack_trace_recorder", None)
    if stack_trace_recorder is None:
        stack_trace_recorder = _StackTraceRecorder()
        _local_data.stack_trace_recorder = stack_trace_recorder
    return stack_trace_recorder.get_stack_trace(
        excluded_modules=config["HIDE_IN_STACKTRACES"],
        include_locals=config["ENABLE_STACKTRACES_LOCALS"],
        skip=skip,
    )


def clear_stack_trace_caches():
    if hasattr(_local_data, "stack_trace_recorder"):
        del _local_data.stack_trace_recorder


_HTML_TYPES = ("text/html", "application/xhtml+xml")


def is_processable_html_response(response):
    content_encoding = response.get("Content-Encoding", "")
    content_type = response.get("Content-Type", "").split(";")[0]
    return (
        not getattr(response, "streaming", False)
        and content_encoding == ""
        and content_type in _HTML_TYPES
    )


def get_csp_nonce(request) -> str | None:
    """
    Retrieve the Content Security Policy nonce from a request if there is one.

    This supports both the django-csp and the built-in Django implementations.
    """
    # django-csp uses request.csp_nonce
    csp_nonce = getattr(request, "csp_nonce", None)
    if csp_nonce is not None:
        return csp_nonce
    # Django's built-in CSP support uses get_nonce(request)
    return compat.get_nonce(request)



import unittest

from debug_toolbar._compat import django_has_tasks_support, task
from debug_toolbar.panels.tasks import TasksPanel

from ..base import BaseTestCase


@task
def sample_task(x, y=1):
    return x + y


class TasksPanelTestCase(BaseTestCase):
    panel_id = TasksPanel.panel_id

    @unittest.skipUnless(not django_has_tasks_support, "Requires Django < 6.0")
    def test_nav_subtitle_without_tasks_support(self):
        """
        On Django < 6.0, django.tasks doesn't exist, so the panel should
        explain that instead of showing task data.
        """
        self.panel.generate_stats(self.request, None)
        self.assertEqual(str(self.panel.nav_subtitle), "Requires Django 6.0+")
        stats = self.panel.get_stats()
        self.assertEqual(stats, {"tasks_available": False, "tasks": []})

    @unittest.skipUnless(django_has_tasks_support, "Requires Django 6.0+")
    def test_no_tasks_queued(self):
        self.panel.generate_stats(self.request, None)
        stats = self.panel.get_stats()
        self.assertEqual(stats, {"tasks_available": True, "tasks": []})
        self.assertEqual(str(self.panel.nav_subtitle), "0 tasks")

    @unittest.skipUnless(django_has_tasks_support, "Requires Django 6.0+")
    def test_records_queued_task(self):
        sample_task.enqueue(2, y=3)

        self.panel.generate_stats(self.request, None)
        stats = self.panel.get_stats()

        self.assertEqual(stats.keys(), {"tasks_available", "tasks"})
        self.assertEqual(stats["tasks_available"], True)
        self.assertEqual(len(stats["tasks"]), 1)
        task_result = stats["tasks"][0]
        self.assertEqual(task_result["task"]["module_path"], f"{__name__}.sample_task")
        self.assertEqual(task_result["task"]["queue_name"], "default")
        self.assertEqual(task_result["task"]["priority"], 0)
        self.assertEqual(task_result["backend"], "default")
        self.assertEqual(task_result["task"]["run_after"], None)
        self.assertEqual(task_result["args"], [2])
        self.assertEqual(task_result["kwargs"], {"y": 3})
        self.assertEqual(task_result["status"], "SUCCESSFUL")

    @unittest.skipUnless(django_has_tasks_support, "Requires Django 6.0+")
    def test_records_queued_task_rendered_in_template(self):
        """Test the Tasks panel's rendered html.

        Since the template for the panel contains the logic for which properties
        are accessed to render the content, we should validate the properties
        explicitly rather than solely relying on test_records_queued_task
        confirming the shape.
        """
        sample_task.enqueue(2, y=3)

        self.panel.generate_stats(self.request, None)
        # This round-trips through the store rather than reading the
        # panel's in-memory stats.
        self.reload_stats()
        content = self.panel.content

        # task.task.module_path
        self.assertIn(f"{__name__}.sample_task", content)
        # task.task.queue_name / task.backend
        self.assertIn("default", content)
        # task.task.priority
        self.assertIn("0", content)
        # task.status
        self.assertIn("SUCCESSFUL", content)
        # task.args
        self.assertIn("[2]", content)
        # task.kwargs
        self.assertIn("{&#x27;y&#x27;: 3}", content)
        self.assertValidHTML(content)

    @unittest.skipUnless(django_has_tasks_support, "Requires Django 6.0+")
    def test_nav_subtitle_counts_multiple_tasks(self):
        sample_task.enqueue(1)
        sample_task.enqueue(2)

        self.panel.generate_stats(self.request, None)

        self.assertEqual(str(self.panel.nav_subtitle), "2 tasks")

    @unittest.skipUnless(django_has_tasks_support, "Requires Django 6.0+")
    def test_disable_instrumentation_stops_recording(self):
        self.panel.disable_instrumentation()
        try:
            sample_task.enqueue(1)
        finally:
            # Restore instrumentation so tearDown's disable call is a no-op
            # rather than raising for double-disconnecting.
            self.panel.enable_instrumentation()

        self.panel.generate_stats(self.request, None)
        stats = self.panel.get_stats()
        self.assertEqual(stats["tasks"], [])



from django.test import SimpleTestCase, override_settings
from django.urls import reverse

from debug_toolbar.panels.versions import VersionsPanel
from debug_toolbar.store import get_store
from tests.test_integration import toolbar_request_id


@override_settings(
    DEBUG=True,
    MIDDLEWARE=[
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.auth.middleware.LoginRequiredMiddleware",
        "debug_toolbar.middleware.DebugToolbarMiddleware",
    ],
)
class LoginNotRequiredTestCase(SimpleTestCase):
    def test_panels(self):
        for uri in (
            "history_sidebar",
            "history_refresh",
            "sql_select",
            "sql_explain",
            "sql_profile",
            "template_source",
        ):
            with self.subTest(uri=uri):
                response = self.client.get(reverse(f"djdt:{uri}"))
                self.assertNotEqual(response.status_code, 200)

    def test_render_panel(self):
        request_id = toolbar_request_id()
        get_store().save_panel(
            request_id, VersionsPanel.panel_id, {"value": "Test data"}
        )
        data = {"request_id": request_id, "panel_id": VersionsPanel.panel_id}

        response = self.client.get(reverse("djdt:render_panel"), query_params=data)
        self.assertEqual(response.status_code, 200)
